from yowsup.structs import ProtocolEntity, ProtocolTreeNode
from .notification import NotificationProtocolEntity
class PictureNotificationProtocolEntity(NotificationProtocolEntity):
    '''
    <notification offline="0" id="{{NOTIFICATION_ID}}" notify="{{NOTIFY_NAME}}" type="picture" 
            t="{{TIMESTAMP}}" from="{{SENDER_JID}}">
    </notification>
    '''

    def __init__(self, _id,  _from, status, timestamp, notify, offline, setJid, setId):
        super(PictureNotificationProtocolEntity, self).__init__("picture", _id, _from, timestamp, notify, offline)
        self.setData(setJid, setId)

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = NotificationProtocolEntity.fromProtocolTreeNode(node)
        entity.__class__ = PictureNotificationProtocolEntity
        return entity


from yowsup.layers.protocol_contacts.protocolentities import RemoveContactNotificationProtocolEntity
from yowsup.structs.protocolentity import ProtocolEntityTest
import time
import unittest

entity = RemoveContactNotificationProtocolEntity("1234", "jid@s.whatsapp.net",
                                                 int(time.time()), "notify", False, "contactjid@s.whatsapp.net")

class RemoveContactNotificationProtocolEntityTest(ProtocolEntityTest, unittest.TestCase):
    def setUp(self):
        self.ProtocolEntity = RemoveContactNotificationProtocolEntity
        self.node = entity.toProtocolTreeNode()



from yowsup.layers.protocol_presence.protocolentities.presence_available import AvailablePresenceProtocolEntity
from yowsup.layers.protocol_presence.protocolentities.test_presence import PresenceProtocolEntityTest

class AvailablePresenceProtocolEntityTest(PresenceProtocolEntityTest):
    def setUp(self):
        super(AvailablePresenceProtocolEntityTest, self).setUp()
        self.ProtocolEntity = AvailablePresenceProtocolEntity
        self.node.setAttribute("type", "available")



from yowsup.layers.protocol_notifications.protocolentities.notification_picture_set import SetPictureNotificationProtocolEntity
from yowsup.structs import ProtocolTreeNode
from yowsup.layers.protocol_notifications.protocolentities.test_notification_picture import PictureNotificationProtocolEntityTest

class SetPictureNotificationProtocolEntityTest(PictureNotificationProtocolEntityTest):
    def setUp(self):
        super(SetPictureNotificationProtocolEntityTest, self).setUp()
        self.ProtocolEntity = SetPictureNotificationProtocolEntity
        setNode = ProtocolTreeNode("set", {"jid": "SET_JID", "id": "123"}, None, None)
        self.node.addChild(setNode)



from yowsup.layers.protocol_iq.protocolentities.test_iq import IqProtocolEntityTest
from yowsup.layers.protocol_profiles.protocolentities import SetStatusIqProtocolEntity
from yowsup.structs import ProtocolTreeNode

class SetStatusIqProtocolEntityTest(IqProtocolEntityTest):
    def setUp(self):
        super(SetStatusIqProtocolEntityTest, self).setUp()
        self.ProtocolEntity = SetStatusIqProtocolEntity
        statusNode = ProtocolTreeNode("status", {}, [], "Hey there, I'm using WhatsApp")
        self.node.addChild(statusNode)



from yowsup.layers import YowProtocolLayer
from .protocolentities import *
from yowsup.layers.protocol_acks.protocolentities import OutgoingAckProtocolEntity
from yowsup.layers.protocol_receipts.protocolentities import OutgoingReceiptProtocolEntity
class YowCallsProtocolLayer(YowProtocolLayer):

    def __init__(self):
        handleMap = {
            "call": (self.recvCall, self.sendCall)
        }
        super(YowCallsProtocolLayer, self).__init__(handleMap)

    def __str__(self):
        return "call Layer"

    def sendCall(self, entity):
        if entity.getTag() == "call":
            self.toLower(entity.toProtocolTreeNode())

    def recvCall(self, node):
        entity = CallProtocolEntity.fromProtocolTreeNode(node)
        if entity.getType() == "offer":
            receipt = OutgoingReceiptProtocolEntity(node["id"], node["from"], callId = entity.getCallId())
            self.toLower(receipt.toProtocolTreeNode())
        else:
            ack = OutgoingAckProtocolEntity(node["id"], "call", None, node["from"])
            self.toLower(ack.toProtocolTreeNode())
        self.toUpper(entity)




from yowsup.layers.auth.protocolentities import AuthProtocolEntity
from yowsup.structs import ProtocolTreeNode
from yowsup.structs.protocolentity import ProtocolEntityTest
import unittest

class AuthProtocolEntityTest(ProtocolEntityTest, unittest.TestCase):
    def setUp(self):
        self.ProtocolEntity = AuthProtocolEntity 
        self.node = ProtocolTreeNode("auth", {"user": "testuser", "mechanism": "WAUTH-2", "passive": "false"})


from yowsup.structs import ProtocolTreeNode
from yowsup.layers.protocol_iq.protocolentities import ResultIqProtocolEntity

'''
<iq type="result" from="{{JID}}@s.whatsapp.net" id="{{IQ:ID}}">
<privacy>
<category name="last" value="all">
</category>
<category name="status" value="all">
</category>
<category name="profile" value="all">
</category>
</privacy>
</iq>
'''

class ResultPrivacyIqProtocolEntity(ResultIqProtocolEntity):
    NODE_PRIVACY="privacy"

    def __init__(self, privacy):
        super(ResultPrivacyIqProtocolEntity, self).__init__()
        self.setProps(privacy)

    def setProps(self, privacy):
        assert type(privacy) is dict, "Privacy must be a dict {name => value}"
        self.privacy = privacy

    def __str__(self):
        out = super(ResultPrivacyIqProtocolEntity, self).__str__()
        out += "Privacy settings\n"
        for name, value in self.privacy.items():
            out += "Category %s  --> %s\n" % (name, value)
        return out

    def toProtocolTreeNode(self):
        node = super(ResultPrivacyIqProtocolEntity, self).toProtocolTreeNode()
        queryNode = ProtocolTreeNode(self.__class__.NODE_PRIVACY)
        node.addChild(queryNode)
        return node

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = super(ResultPrivacyIqProtocolEntity, ResultPrivacyIqProtocolEntity).fromProtocolTreeNode(node)
        entity.__class__ = ResultPrivacyIqProtocolEntity
        privacyNode = node.getChild(ResultPrivacyIqProtocolEntity.NODE_PRIVACY)
        privacy = {}
        for categoryNode in privacyNode.getAllChildren():
            privacy[categoryNode["name"]] = categoryNode["value"]
        entity.setProps(privacy)
        return entity



from yowsup.stacks import  YowStackBuilder
from .layer import YowsupCliLayer
from yowsup.layers.auth import AuthError
from yowsup.layers import YowLayerEvent
from yowsup.layers.auth import YowAuthenticationProtocolLayer
import sys

class YowsupCliStack(object):
    def __init__(self, credentials, encryptionEnabled = True):
        stackBuilder = YowStackBuilder()

        self.stack = stackBuilder\
            .pushDefaultLayers(encryptionEnabled)\
            .push(YowsupCliLayer)\
            .build()

        # self.stack.setCredentials(credentials)
        self.stack.setCredentials(credentials)

    def start(self):
        print("Yowsup Cli client\n==================\nType /help for available commands\n")
        self.stack.broadcastEvent(YowLayerEvent(YowsupCliLayer.EVENT_START))

        try:
            self.stack.loop(timeout = 0.5, discrete = 0.5)
        except AuthError as e:
            print("Auth Error, reason %s" % e)
        except KeyboardInterrupt:
            print("\nYowsdown")
            sys.exit(0)



from yowsup.layers import YowLayer, YowLayerEvent
from yowsup.layers.protocol_iq.protocolentities import IqProtocolEntity
from yowsup.layers.network import YowNetworkLayer
from yowsup.layers.auth import YowAuthenticationProtocolLayer
from yowsup.layers.protocol_receipts.protocolentities import OutgoingReceiptProtocolEntity
from yowsup.layers.protocol_acks.protocolentities import IncomingAckProtocolEntity
import inspect

class ProtocolEntityCallback(object):
    def __init__(self, entityType):
        self.entityType = entityType

    def __call__(self, fn):
        fn.entity_callback = self.entityType
        return fn
    

class YowInterfaceLayer(YowLayer):

    def __init__(self):
        super(YowInterfaceLayer, self).__init__()
        self.entity_callbacks = {}
        self.iqRegistry = {}
        # self.receiptsRegistry = {}
        members = inspect.getmembers(self, predicate=inspect.ismethod)
        for m in members:
            if hasattr(m[1], "entity_callback"):
                fname = m[0]
                fn = m[1]
                self.entity_callbacks[fn.entity_callback] = getattr(self, fname)
            

    def _sendIq(self, iqEntity, onSuccess = None, onError = None):
        assert iqEntity.getTag() == "iq", "Expected *IqProtocolEntity in _sendIq, got %s" % iqEntity.getTag()
        self.iqRegistry[iqEntity.getId()] = (iqEntity, onSuccess, onError)
        self.toLower(iqEntity)

    # def _sendReceipt(self, outgoingReceiptProtocolEntity, onAck = None):
    #     assert outgoingReceiptProtocolEntity.__class__ == OutgoingReceiptProtocolEntity,\
    #         "Excepted OutgoingReceiptProtocolEntity in _sendReceipt, got %s" % outgoingReceiptProtocolEntity.__class__
    #     self.receiptsRegistry[outgoingReceiptProtocolEntity.getId()] = (outgoingReceiptProtocolEntity, onAck)
    #     self.toLower(outgoingReceiptProtocolEntity)

    # def processReceiptsRegistry(self, incomingAckProtocolEntity):
    #     '''
    #     entity: IncomingAckProtocolEntity
    #     '''
    #
    #     if incomingAckProtocolEntity.__class__ != IncomingAckProtocolEntity:
    #         return False
    #
    #     receipt_id = incomingAckProtocolEntity.getId()
    #     if receipt_id in self.receiptsRegistry:
    #         originalReceiptEntity, ackClbk = self.receiptsRegistry[receipt_id]
    #         del self.receiptsRegistry[receipt_id]
    #
    #         if ackClbk:
    #             ackClbk(incomingAckProtocolEntity, originalReceiptEntity)
    #
    #         return True
    #
    #     return False


    def processIqRegistry(self, entity):
        """
        :type entity: IqProtocolEntity
        """
        if entity.getTag() == "iq":
            iq_id = entity.getId()
            if iq_id in self.iqRegistry:
                originalIq, successClbk, errorClbk = self.iqRegistry[iq_id]
                del self.iqRegistry[iq_id]

                if entity.getType() ==  IqProtocolEntity.TYPE_RESULT and successClbk:
                    successClbk(entity, originalIq)
                elif entity.getType() == IqProtocolEntity.TYPE_ERROR and errorClbk:
                    errorClbk(entity, originalIq)
                return True

        return False

    def getOwnJid(self, full = True):
        return self.getLayerInterface(YowAuthenticationProtocolLayer).getUsername(full)

    def connect(self):
        self.getLayerInterface(YowNetworkLayer).connect()

    def disconnect(self):
        disconnectEvent = YowLayerEvent(YowNetworkLayer.EVENT_STATE_DISCONNECT)
        self.broadcastEvent(disconnectEvent)

    def send(self, data):
        self.toLower(data)

    def receive(self, entity):
        if not self.processIqRegistry(entity):
            entityType = entity.getTag()
            if entityType in self.entity_callbacks:
                self.entity_callbacks[entityType](entity)
            else:
                self.toUpper(entity)
 
    def __str__(self):
        return "Interface Layer"



from yowsup.structs import ProtocolTreeNode
from .iq_sync import SyncIqProtocolEntity

class ResultSyncIqProtocolEntity(SyncIqProtocolEntity):
    '''
    <iq type="result" from="491632092557@s.whatsapp.net" id="1417046561-4">
    <sync index="0" wait="166952" last="true" version="1417046548593182" sid="1.30615237617e+17">
        <in>
            <user jid="{{jid}}>{{number}}</user>
        </in>
        <out>
            <user jid="{{jid}}">
            {{number}}
        </user>
        </out>
        <invalid>
            <user>
                abcdefgh
            </user>
        </invalid>

    </sync>
    </iq>
    '''

    def __init__(self,_id, sid, index, last, version, inNumbers, outNumbers, invalidNumbers, wait = None):
        super(ResultSyncIqProtocolEntity, self).__init__("result", _id, sid, index, last)
        self.setResultSyncProps(version, inNumbers, outNumbers, invalidNumbers, wait)

    def setResultSyncProps(self, version, inNumbers, outNumbers, invalidNumbers, wait = None):
        assert type(inNumbers) is dict, "in numbers must be a dict {number -> jid}"
        assert type(outNumbers) is dict, "out numbers must be a dict {number -> jid}"
        assert type(invalidNumbers) is list, "invalid numbers must be a list"

        self.inNumbers = inNumbers
        self.outNumbers = outNumbers
        self.invalidNumbers = invalidNumbers
        self.wait = int(wait) if wait is not None else None
        self.version = version


    def __str__(self):
        out  = super(SyncIqProtocolEntity, self).__str__()
        if self.wait is not None:
            out += "Wait: %s\n" % self.wait
        out += "Version: %s\n" % self.version
        out += "In Numbers: %s\n" % (",".join(self.inNumbers))
        out += "Out Numbers: %s\n" % (",".join(self.outNumbers))
        out += "Invalid Numbers: %s\n" % (",".join(self.invalidNumbers))

        return out

    def toProtocolTreeNode(self):
        
        outUsers = [ProtocolTreeNode("user", {"jid": jid}, None, number) for number, jid in self.outNumbers.items()]
        inUsers = [ProtocolTreeNode("user", {"jid": jid}, None, number) for number, jid in self.inNumbers.items()]
        invalidUsers = [ProtocolTreeNode("user", {}, None, number) for number in self.invalidNumbers]

        node = super(ResultSyncIqProtocolEntity, self).toProtocolTreeNode()
        syncNode = node.getChild("sync")
        syncNode.setAttribute("version", self.version)

        if self.wait is not None:
            syncNode.setAttribute("wait", str(self.wait))

        if len(outUsers):
            syncNode.addChild(ProtocolTreeNode("out", children = outUsers))

        if len(inUsers):
            syncNode.addChild(ProtocolTreeNode("in", children = inUsers))

        if len(invalidUsers):
            syncNode.addChildren([ProtocolTreeNode("invalid", children = invalidUsers)])

        return node

    @staticmethod
    def fromProtocolTreeNode(node):
        syncNode         = node.getChild("sync")
        outNode          = syncNode.getChild("out")
        inNode           = syncNode.getChild("in")
        invalidNode      = syncNode.getChild("invalid")
        outUsers         = outNode.getAllChildren() if outNode else []
        inUsers          = inNode.getAllChildren()  if inNode else []
        invalidUsers     = [inode.data for inode in invalidNode.getAllChildren()] if invalidNode else []

        outUsersDict = {}
        for u in outUsers:
            outUsersDict[u.data] = u.getAttributeValue("jid")

        inUsersDict = {}
        for u in inUsers:
            inUsersDict[u.data] = u.getAttributeValue("jid")

        entity           = SyncIqProtocolEntity.fromProtocolTreeNode(node)
        entity.__class__ = ResultSyncIqProtocolEntity

        entity.setResultSyncProps(syncNode.getAttributeValue("version"),
            inUsersDict,
            outUsersDict,
            invalidUsers,
            syncNode.getAttributeValue("wait")
            )
   
        return entity



from .iq_groups_create      import CreateGroupsIqProtocolEntity
from .iq_groups_create_success import SuccessCreateGroupsIqProtocolEntity
from .iq_groups_leave       import LeaveGroupsIqProtocolEntity
from .iq_groups_leave_success import SuccessLeaveGroupsIqProtocolEntity
from .iq_groups_list        import ListGroupsIqProtocolEntity
from .iq_groups_info        import InfoGroupsIqProtocolEntity
from .iq_groups_subject     import SubjectGroupsIqProtocolEntity
from .iq_groups_participants import ParticipantsGroupsIqProtocolEntity
from .iq_groups_participants_add import AddParticipantsIqProtocolEntity
from .iq_groups_participants_promote import PromoteParticipantsIqProtocolEntity
from .iq_groups_participants_demote import DemoteParticipantsIqProtocolEntity
from .iq_groups_participants_add_success import SuccessAddParticipantsIqProtocolEntity
from .iq_groups_participants_add_failure import FailureAddParticipantsIqProtocolEntity
from .iq_groups_participants_remove import RemoveParticipantsIqProtocolEntity
from .iq_groups_participants_remove_success import SuccessRemoveParticipantsIqProtocolEntity
from .iq_result_groups_list import ListGroupsResultIqProtocolEntity
from .iq_result_participants_list import ListParticipantsResultIqProtocolEntity
from .iq_result_groups_info import InfoGroupsResultIqProtocolEntity
from .notification_groups    import GroupsNotificationProtocolEntity
from .notification_groups_subject  import SubjectGroupsNotificationProtocolEntity
from .notification_groups_create import CreateGroupsNotificationProtocolEntity
from .notification_groups_add import AddGroupsNotificationProtocolEntity
from .notification_groups_remove import RemoveGroupsNotificationProtocolEntity



from yowsup.common import YowConstants
from yowsup.layers import YowLayerEvent, YowProtocolLayer, EventCallback
from .keystream import KeyStream
from yowsup.common.tools import TimeTools
from .layer_crypt import YowCryptLayer
from yowsup.layers.network import YowNetworkLayer
from .autherror import AuthError
from .protocolentities import *
from yowsup.common.tools import StorageTools
from .layer_interface_authentication import YowAuthenticationProtocolLayerInterface
import base64
class YowAuthenticationProtocolLayer(YowProtocolLayer):
    EVENT_LOGIN      = "org.openwhatsapp.yowsup.event.auth.login"
    EVENT_AUTHED  = "org.openwhatsapp.yowsup.event.auth.authed"
    PROP_CREDENTIALS = "org.openwhatsapp.yowsup.prop.auth.credentials"
    PROP_PASSIVE = "org.openwhatsapp.yowsup.prop.auth.passive"

    def __init__(self):
        handleMap = {
            "stream:features": (self.handleStreamFeatures, None),
            "failure": (self.handleFailure, None),
            "success": (self.handleSuccess, None),
            "challenge": (self.handleChallenge, None),
            "stream:error": (self.handleStreamError, None),
        }
        super(YowAuthenticationProtocolLayer, self).__init__(handleMap)
        self.interface = YowAuthenticationProtocolLayerInterface(self)
        self.credentials = None #left for backwards-compat
        self._credentials = None #new style set

    def __str__(self):
        return "Authentication Layer"

    def __getCredentials(self, credentials = None):
        u, pb64 = credentials or self.getProp(YowAuthenticationProtocolLayer.PROP_CREDENTIALS)
        if type(pb64) is str:
            pb64 = pb64.encode()
        password = base64.b64decode(pb64)
        return (u, bytearray(password))

    def setCredentials(self, credentials):
        self.setProp(YowAuthenticationProtocolLayer.PROP_CREDENTIALS, credentials) #keep for now
        self._credentials = self.__getCredentials(credentials)

    def getUsername(self, full = False):
        if self._credentials:
            return self._credentials[0] if not full else ("%s@%s" % (self._credentials[0], YowConstants.WHATSAPP_SERVER))
        else:
            prop = self.getProp(YowAuthenticationProtocolLayer.PROP_CREDENTIALS)
            return prop[0] if prop else None
        
    @EventCallback(YowNetworkLayer.EVENT_STATE_CONNECTED)
    def onConnected(self, yowLayerEvent):
        self.login()

    ## general methods
    def login(self):
        self.credentials = self._credentials or self.__getCredentials()
        self._sendFeatures()
        self._sendAuth()

    ###recieved node handlers handlers
    def handleStreamFeatures(self, node):
        nodeEntity = StreamFeaturesProtocolEntity.fromProtocolTreeNode(node)
        self.toUpper(nodeEntity)

    def handleSuccess(self, node):
        if(node.data != None): StorageTools.writeNonce(self.credentials[0],node.data)
        successEvent = YowLayerEvent(self.__class__.EVENT_AUTHED, passive = self.getProp(self.__class__.PROP_PASSIVE))
        self.broadcastEvent(successEvent)
        nodeEntity = SuccessProtocolEntity.fromProtocolTreeNode(node)
        self.toUpper(nodeEntity)

    def handleFailure(self, node):
        nodeEntity = FailureProtocolEntity.fromProtocolTreeNode(node)
        self.toUpper(nodeEntity)
        self.broadcastEvent(YowLayerEvent(YowNetworkLayer.EVENT_STATE_DISCONNECT, reason = "Authentication Failure"))
        raise AuthError(nodeEntity.getReason())

    def handleChallenge(self, node):
        nodeEntity = ChallengeProtocolEntity.fromProtocolTreeNode(node)
        self._sendResponse(nodeEntity.getNonce())

    def handleStreamError(self, node):
        if node.getChild("text"):
            nodeEntity = StreamErrorConflictProtocolEntity.fromProtocolTreeNode(node)
        elif node.getChild("ack"):
            nodeEntity = StreamErrorAckProtocolEntity.fromProtocolTreeNode(node)
        else:
            raise AuthError("Unhandled stream:error node:\n%s" % node)
        self.toUpper(nodeEntity)

    ##senders
    def _sendFeatures(self):
        self.entityToLower(StreamFeaturesProtocolEntity(["readreceipts", "groups_v2", "privacy", "presence"]))

    def _sendAuth(self):
        passive = self.getProp(self.__class__.PROP_PASSIVE, False)
        nonce = StorageTools.getNonce(self.credentials[0])

        if nonce == None:
            self.entityToLower(AuthProtocolEntity(self.credentials[0], passive=passive))
        else:
            inputKey, outputKey, authBlob = self.generateAuthBlob(nonce)
            #to prevent enr whole response
            self.broadcastEvent(YowLayerEvent(YowCryptLayer.EVENT_KEYS_READY, keys = (inputKey, None)))
            self.entityToLower(AuthProtocolEntity(self.credentials[0], passive=passive, nonce=authBlob))
            self.broadcastEvent(YowLayerEvent(YowCryptLayer.EVENT_KEYS_READY, keys = (inputKey, outputKey)))


    def _sendResponse(self,nonce):
        inputKey, outputKey, authBlob = self.generateAuthBlob(nonce)
        responseEntity = ResponseProtocolEntity(authBlob)

        #to prevent enr whole response
        self.broadcastEvent(YowLayerEvent(YowCryptLayer.EVENT_KEYS_READY, keys = (inputKey, None)))
        self.entityToLower(responseEntity)
        self.broadcastEvent(YowLayerEvent(YowCryptLayer.EVENT_KEYS_READY, keys = (inputKey, outputKey)))
        #YowCryptLayer.setProp("outputKey", outputKey)


    def generateAuthBlob(self, nonce):
        keys = KeyStream.generateKeys(self.credentials[1], nonce)

        inputKey = KeyStream(keys[2], keys[3])
        outputKey = KeyStream(keys[0], keys[1])

        #YowCryptLayer.setProp("inputKey", inputKey)


        nums = bytearray(4)

        #nums = [0] * 4


        username_bytes = list(map(ord, self.credentials[0]))
        nums.extend(username_bytes)
        nums.extend(nonce)

        utcNow = str(int(TimeTools.utcTimestamp()))

        time_bytes =  list(map(ord, utcNow))

        nums.extend(time_bytes)

        encoded = outputKey.encodeMessage(nums, 0, 4, len(nums) - 4)
        authBlob = "".join(map(chr, encoded))

        return (inputKey, outputKey, authBlob)



from yowsup.layers.auth.protocolentities.stream_features import StreamFeaturesProtocolEntity
from yowsup.structs import ProtocolTreeNode
from yowsup.structs.protocolentity import ProtocolEntityTest
import unittest

class StreamFeaturesProtocolEntityTest(ProtocolEntityTest, unittest.TestCase):
    def setUp(self):
        self.ProtocolEntity = StreamFeaturesProtocolEntity
        attribs = {
        }
        childNode = ProtocolTreeNode("receipt_acks", {}, None, None)
        self.node = ProtocolTreeNode("stream:features", attribs, [childNode])


from yowsup.layers.protocol_iq.protocolentities import IqProtocolEntity
from yowsup.structs import ProtocolTreeNode

'''
<iq xmlns="privacy" type="get" id="{{IQ_ID}}">
<privacy>
</privacy>
</iq>
'''

class GetPrivacyIqProtocolEntity(IqProtocolEntity):
    XMLNS = "privacy"
    def __init__(self):
        super(GetPrivacyIqProtocolEntity, self).__init__(self.__class__.XMLNS, _type="get")

    def toProtocolTreeNode(self):
        node = super(GetPrivacyIqProtocolEntity, self).toProtocolTreeNode()
        queryNode = ProtocolTreeNode(self.__class__.XMLNS)
        node.addChild(queryNode)
        return node

    @staticmethod
    def fromProtocolTreeNode(node):
        assert node.getChild(GetPrivacyIqProtocolEntity.XMLNS) is not None, "Not a get privacy iq node %s" % node
        entity = IqProtocolEntity.fromProtocolTreeNode(node)
        entity.__class__ = GetPrivacyIqProtocolEntity
        return entity



from yowsup.layers.protocol_chatstate.protocolentities.chatstate_outgoing import OutgoingChatstateProtocolEntity
from yowsup.structs.protocolentity import ProtocolEntityTest
import unittest

entity = OutgoingChatstateProtocolEntity(OutgoingChatstateProtocolEntity.STATE_PAUSED, "jid@s.whatsapp.net")

class OutgoingChatstateProtocolEntityTest(ProtocolEntityTest, unittest.TestCase):
    def setUp(self):
        self.ProtocolEntity = OutgoingChatstateProtocolEntity
        self.node = entity.toProtocolTreeNode()



from yowsup.layers.protocol_messages.protocolentities.message_text import TextMessageProtocolEntity
from yowsup.structs import ProtocolTreeNode
from yowsup.layers.protocol_messages.protocolentities.test_message import MessageProtocolEntityTest

class TextMessageProtocolEntityTest(MessageProtocolEntityTest):
    def setUp(self):
        super(TextMessageProtocolEntityTest, self).setUp()
        self.ProtocolEntity = TextMessageProtocolEntity
        bodyNode = ProtocolTreeNode("body", {}, None, "body_data")
        self.node.addChild(bodyNode)



import unittest
from yowsup.layers.coder.decoder import ReadDecoder
from yowsup.layers.coder.tokendictionary import TokenDictionary
from yowsup.structs import ProtocolTreeNode
class DecoderTest(unittest.TestCase):
    def setUp(self):
        self.decoder = ReadDecoder(TokenDictionary())
        self.decoder.streamStarted = True

    def test_decode(self):
        data = [248, 6, 89, 165, 252, 3, 120, 121, 122, 252, 4, 102, 111, 114, 109, 252, 3, 97, 98, 99, 248, 1,
                                  248, 4, 87, 236, 99, 252, 3, 49, 50, 51, 252, 6, 49, 50, 51, 52, 53, 54]
        node = self.decoder.getProtocolTreeNode(data)
        targetNode = ProtocolTreeNode("message", {"form": "abc", "to":"xyz"}, [ProtocolTreeNode("media", {"width" : "123"}, data="123456")])
        self.assertEqual(node, targetNode)



from .env import YowsupEnv
import base64
import hashlib


class AndroidYowsupEnv(YowsupEnv):
    _SIGNATURE = "MIIDMjCCAvCgAwIBAgIETCU2pDALBgcqhkjOOAQDBQAwfDELMAkGA1UEBhMCVVMxEzARBgNVBAgTCkNhbGlmb3JuaWExFDASBgNV" \
        "BAcTC1NhbnRhIENsYXJhMRYwFAYDVQQKEw1XaGF0c0FwcCBJbmMuMRQwEgYDVQQLEwtFbmdpbmVlcmluZzEUMBIGA1UEAxMLQnJ" \
        "pYW4gQWN0b24wHhcNMTAwNjI1MjMwNzE2WhcNNDQwMjE1MjMwNzE2WjB8MQswCQYDVQQGEwJVUzETMBEGA1UECBMKQ2FsaWZvcm5" \
        "pYTEUMBIGA1UEBxMLU2FudGEgQ2xhcmExFjAUBgNVBAoTDVdoYXRzQXBwIEluYy4xFDASBgNVBAsTC0VuZ2luZWVyaW5nMRQwEg" \
        "YDVQQDEwtCcmlhbiBBY3RvbjCCAbgwggEsBgcqhkjOOAQBMIIBHwKBgQD9f1OBHXUSKVLfSpwu7OTn9hG3UjzvRADDHj+AtlEm" \
        "aUVdQCJR+1k9jVj6v8X1ujD2y5tVbNeBO4AdNG/yZmC3a5lQpaSfn+gEexAiwk+7qdf+t8Yb+DtX58aophUPBPuD9tPFHsMCN" \
        "VQTWhaRMvZ1864rYdcq7/IiAxmd0UgBxwIVAJdgUI8VIwvMspK5gqLrhAvwWBz1AoGBAPfhoIXWmz3ey7yrXDa4V7l5lK+7+jr" \
        "qgvlXTAs9B4JnUVlXjrrUWU/mcQcQgYC0SRZxI+hMKBYTt88JMozIpuE8FnqLVHyNKOCjrh4rs6Z1kW6jfwv6ITVi8ftiegEkO" \
        "8yk8b6oUZCJqIPf4VrlnwaSi2ZegHtVJWQBTDv+z0kqA4GFAAKBgQDRGYtLgWh7zyRtQainJfCpiaUbzjJuhMgo4fVWZIvXHaS" \
        "HBU1t5w//S0lDK2hiqkj8KpMWGywVov9eZxZy37V26dEqr/c2m5qZ0E+ynSu7sqUD7kGx/zeIcGT0H+KAVgkGNQCo5Uc0koLRW" \
        "YHNtYoIvt5R3X6YZylbPftF/8ayWTALBgcqhkjOOAQDBQADLwAwLAIUAKYCp0d6z4QQdyN74JDfQ2WCyi8CFDUM4CaNB+ceVXd" \
        "KtOrNTQcc0e+t"

    _MD5_CLASSES = "7UDPOXwpiLBvEjT8uNwsuA=="
    _KEY = "eQV5aq/Cg63Gsq1sshN9T3gh+UUp0wIw0xgHYT1bnCjEqOJQKCRrWxdAe2yvsDeCJL+Y4G3PRD2HUF7oUgiGo8vGlNJOaux26k+A2F3hj8A="

    _VERSION = "2.12.440"
    _OS_NAME = "Android"
    _OS_VERSION = "4.3"
    _DEVICE_NAME = "GalaxyS3"
    _AXOLOTL = True

    def getVersion(self):
        return self.__class__._VERSION

    def getOSName(self):
        return self.__class__._OS_NAME

    def getOSVersion(self):
        return self.__class__._OS_VERSION

    def getDeviceName(self):
        return self.__class__._DEVICE_NAME

    def isAxolotlEnabled(self):
        return self.__class__._AXOLOTL

    def getToken(self, phoneNumber):
        keyDecoded = bytearray(base64.b64decode(self.__class__._KEY))
        sigDecoded = base64.b64decode(self.__class__._SIGNATURE)
        clsDecoded = base64.b64decode(self.__class__._MD5_CLASSES)
        data = sigDecoded + clsDecoded + phoneNumber.encode()

        opad = bytearray()
        ipad = bytearray()
        for i in range(0, 64):
            opad.append(0x5C ^ keyDecoded[i])
            ipad.append(0x36 ^ keyDecoded[i])
        hash = hashlib.sha1()
        subHash = hashlib.sha1()
        try:
            subHash.update(ipad + data)
            hash.update(opad + subHash.digest())
        except TypeError:
            subHash.update(bytes(ipad + data))
            hash.update(bytes(opad + subHash.digest()))
        result = base64.b64encode(hash.digest())
        return result



from yowsup.stacks import  YowStackBuilder
from .layer import SendLayer
from yowsup.layers.auth import AuthError
from yowsup.layers import YowLayerEvent
from yowsup.layers.auth import YowAuthenticationProtocolLayer
from yowsup.layers.network import YowNetworkLayer


class YowsupSendStack(object):
    def __init__(self, credentials, messages, encryptionEnabled = True):
        """
        :param credentials:
        :param messages: list of (jid, message) tuples
        :param encryptionEnabled:
        :return:
        """
        stackBuilder = YowStackBuilder()

        self.stack = stackBuilder\
            .pushDefaultLayers(encryptionEnabled)\
            .push(SendLayer)\
            .build()

        self.stack.setProp(SendLayer.PROP_MESSAGES, messages)
        self.stack.setProp(YowAuthenticationProtocolLayer.PROP_PASSIVE, True)
        self.stack.setCredentials(credentials)

    def start(self):
        self.stack.broadcastEvent(YowLayerEvent(YowNetworkLayer.EVENT_STATE_CONNECT))
        try:
            self.stack.loop()
        except AuthError as e:
            print("Authentication Error: %s" % e.message)



from yowsup.layers.protocol_iq.protocolentities.iq_result import ResultIqProtocolEntity
from yowsup.structs.protocoltreenode import ProtocolTreeNode
class ResultLastseenIqProtocolEntity(ResultIqProtocolEntity):
    def __init__(self, jid, seconds, _id = None):
        super(ResultLastseenIqProtocolEntity, self).__init__(_from=jid, _id=_id)
        self.setSeconds(seconds)

    def setSeconds(self, seconds):
        self.seconds = int(seconds)

    def getSeconds(self):
        return self.seconds

    def __str__(self):
        out = super(ResultIqProtocolEntity, self).__str__()
        out += "Seconds: %s\n" % self.seconds
        return out

    def toProtocolTreeNode(self):
        node = super(ResultLastseenIqProtocolEntity, self).toProtocolTreeNode()
        node.addChild(ProtocolTreeNode("query", {"seconds": str(self.seconds)}))
        return node

    @staticmethod
    def fromProtocolTreeNode(node):
        return ResultLastseenIqProtocolEntity(node["from"], node.getChild("query")["seconds"], node["id"])


from .env import YowsupEnv
import base64
import hashlib
class S40YowsupEnv(YowsupEnv):
    _VERSION = "2.13.39"
    _OS_NAME= "S40"
    _OS_VERSION = "14.26"
    _DEVICE_NAME = "302"
    _MANUFACTURER = "Nokia"
    _TOKEN_STRING  = "PdA2DJyKoUrwLw1Bg6EIhzh502dF9noR9uFCllGk1456529096701{phone}"
    _AXOLOTL = True

    def getVersion(self):
        return self.__class__._VERSION

    def getOSName(self):
        return self.__class__._OS_NAME

    def getOSVersion(self):
        return self.__class__._OS_VERSION

    def getDeviceName(self):
        return self.__class__._DEVICE_NAME

    def isAxolotlEnabled(self):
        return self.__class__._AXOLOTL

    def getToken(self, phoneNumber):
        return hashlib.md5(self.__class__._TOKEN_STRING.format(phone = phoneNumber).encode()).hexdigest()

    def getUserAgent(self):
        return self.__class__._USERAGENT_STRING.format(
            WHATSAPP_VERSION = self.getVersion(),
            OS_NAME = self.getOSName() + "Version",
            OS_VERSION = self.getOSVersion(),
            DEVICE_NAME = self.getDeviceName()
        )



from yowsup.structs import ProtocolEntity, ProtocolTreeNode
from .iq_groups_participants import ParticipantsGroupsIqProtocolEntity

class AddParticipantsIqProtocolEntity(ParticipantsGroupsIqProtocolEntity):
    '''
    <iq type="set" id="{{id}}" xmlns="w:g2", to="{{group_jid}}">
        <add>
            <participant jid="{{jid}}"></participant>
            <participant jid="{{jid}}"></participant>
        </add>
    </iq>
    '''

    def __init__(self, group_jid, participantList, _id = None):
        super(AddParticipantsIqProtocolEntity, self).__init__(group_jid, participantList, "add", _id = _id)

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = super(AddParticipantsIqProtocolEntity, AddParticipantsIqProtocolEntity).fromProtocolTreeNode(node)
        entity.__class__ = AddParticipantsIqProtocolEntity
        participantList = []
        for participantNode in node.getChild("add").getAllChildren():
            participantList.append(participantNode["jid"])
        entity.setProps(node.getAttributeValue("to"), participantList)
        return entity



from yowsup.layers import YowProtocolLayerTest
from yowsup.layers.protocol_acks import YowAckProtocolLayer
from yowsup.layers.protocol_acks.protocolentities.test_ack_incoming import entity as incomingAckEntity
from yowsup.layers.protocol_acks.protocolentities.test_ack_outgoing import entity as outgoingAckEntity
class YowAckProtocolLayerTest(YowProtocolLayerTest, YowAckProtocolLayer):
    def setUp(self):
        YowAckProtocolLayer.__init__(self)

    def test_receive(self):
        self.assertReceived(incomingAckEntity)

    def test_send(self):
        self.assertSent(outgoingAckEntity)



from yowsup.layers.protocol_contacts.protocolentities.iq_sync_result import ResultSyncIqProtocolEntity
from yowsup.structs.protocolentity import ProtocolEntityTest
import unittest

entity = ResultSyncIqProtocolEntity("123", "1.30615237617e+17", 0,
                                            True, "123456", {"12345678": "12345678@s.whatsapp.net"},
                                            {"12345678": "12345678@s.whatsapp.net"}, ["1234"])

class ResultSyncIqProtocolEntityTest(ProtocolEntityTest, unittest.TestCase):
    def setUp(self):
        self.ProtocolEntity = ResultSyncIqProtocolEntity
        self.node = entity.toProtocolTreeNode()

    def test_delta_result(self):
        del self.node.getChild("sync")["wait"]
        self.test_generation()



from yowsup.structs import ProtocolEntity, ProtocolTreeNode
class CallProtocolEntity(ProtocolEntity):
    '''
    <call offline="0" from="{{CALLER_JID}}" id="{{ID}}" t="{{TIMESTAMP}}" notify="{{CALLER_PUSHNAME}}" retry="{{RETRY}}" e="{{?}}">
    </call>
    
    '''
    def __init__(self, _id, _type, timestamp, notify = None, offline = None, retry = None, e = None, callId = None, _from = None, _to = None):
        super(CallProtocolEntity, self).__init__("call")
        self._id            = _id or self._generateId()
        self._type       = _type
        self._from      = _from
        self._to           = _to
        self.timestamp  = int(timestamp)
        self.notify     = notify
        self.offline    = offline == "1"
        self.retry      = retry
        self.e             = e
        self.callId     = callId

    def __str__(self):
        out = "Call\n"
        if self.getFrom() is not None:
            out += "From: %s\n" % self.getFrom()
        if self.getTo() is not None:
            out += "To: %s\n" % self.getTo()
        if self.getType() is not None:
            out += "Type: %s\n" % self.getType()
        if self.getCallId() is not None:
            out += "Call ID: %s\n" % self.getCallId()
        return out

    def getFrom(self, full = True):
        return self._from if full else self._from.split('@')[0]
        
    def getTo(self):
        return self._to

    def getId(self):
        return self._id
        
    def getType(self):
        return self._type
        
    def getCallId(self):
        return self.callId

    def getTimestamp(self):
        return self.timestamp

    def toProtocolTreeNode(self):
        children = []
        attribs = {
            "t"         : str(self.timestamp),
            "offline"   : "1" if self.offline else "0",
            "id"        : self._id,
        }
        if self._from is not None:
            attribs["from"] = self._from
        if self._to is not None:
            attribs["to"] = self._to
        if self.retry is not None:
            attribs["retry"] = self.retry
        if self.e is not None:
            attribs["e"] = self.e
        if self.notify is not None:
            attribs["notify"] = self.notify
        if self._type in ["offer", "transport", "relaylatency", "reject", "terminate"]:
            child = ProtocolTreeNode(self._type, {"call-id": self.callId})
            children.append(child)
        return self._createProtocolTreeNode(attribs, children = children, data = None)

    @staticmethod
    def fromProtocolTreeNode(node):
        (_type, callId) = [None] * 2
        offer = node.getChild("offer")
        transport = node.getChild("transport")
        relaylatency = node.getChild("relaylatency")
        reject = node.getChild("reject")
        terminate = node.getChild("terminate")
        if offer:
            _type = "offer"
            callId = offer.getAttributeValue("call-id")
        elif transport:
            _type = "transport"
            callId = transport.getAttributeValue("call-id")
        elif relaylatency:
            _type = "relaylatency"
            callId = relaylatency.getAttributeValue("call-id")
        elif reject:
            _type = "reject"
            callId = reject.getAttributeValue("call-id")
        elif terminate:
            _type = "terminate"
            callId = terminate.getAttributeValue("call-id")
        return CallProtocolEntity(
            node.getAttributeValue("id"),
            _type,
            node.getAttributeValue("t"),
            node.getAttributeValue("notify"),
            node.getAttributeValue("offline"),
            node.getAttributeValue("retry"),
            node.getAttributeValue("e"),
            callId,
            node.getAttributeValue("from"),
            node.getAttributeValue("to")
            )



from yowsup.structs import ProtocolEntity, ProtocolTreeNode
from .iq import IqProtocolEntity
class PingIqProtocolEntity(IqProtocolEntity):

    '''
    Receive
    <iq type="get" xmlns="urn:xmpp:ping" from="s.whatsapp.net" id="1416174955-ping">
    </iq>
    Send
    <iq type="get" xmlns="w:p" to="s.whatsapp.net" id="1416174955-ping">
    </iq>
    '''

    def __init__(self, _from = None, to = None, _id = None):
        super(PingIqProtocolEntity, self).__init__("urn:xmpp:ping" if _from else "w:p", _id = _id, _type = "get", _from = _from, to = to)



from yowsup.common import YowConstants
from yowsup.layers.protocol_iq.protocolentities.test_iq_result import ResultIqProtocolEntityTest
from yowsup.layers.axolotl.protocolentities import ResultGetKeysIqProtocolEntity
from yowsup.structs import ProtocolTreeNode
from axolotl.util.keyhelper import KeyHelper
from axolotl.ecc.curve import Curve

class ResultGetKeysIqProtocolEntityTest(ResultIqProtocolEntityTest):
    def setUp(self):
        super(ResultIqProtocolEntityTest, self).setUp()
        self.ProtocolEntity = ResultGetKeysIqProtocolEntity
        listNode = ProtocolTreeNode("list")
        self.node.addChild(listNode)

        for i in range(0, 1):
            userNode = ProtocolTreeNode("user", {"jid": "user_%s@%s" % (i, YowConstants.WHATSAPP_SERVER)})
            listNode.addChild(userNode)
            registrationNode = ProtocolTreeNode("registration",
                                                data = ResultGetKeysIqProtocolEntity._intToBytes(
                                                    KeyHelper.generateRegistrationId()))

            typeNode = ProtocolTreeNode("type", data = ResultGetKeysIqProtocolEntity._intToBytes(Curve.DJB_TYPE))

            identityKeyPair = KeyHelper.generateIdentityKeyPair()
            identityNode = ProtocolTreeNode("identity", data=identityKeyPair.getPublicKey().getPublicKey().getPublicKey())

            signedPreKey = KeyHelper.generateSignedPreKey(identityKeyPair, i)

            signedPreKeyNode = ProtocolTreeNode("skey")
            signedPreKeyNode_idNode = ProtocolTreeNode("id",
                                                       data = ResultGetKeysIqProtocolEntity._intToBytes(
                                                           signedPreKey.getId()))

            signedPreKeyNode_valueNode = ProtocolTreeNode("value",
                                                          data = signedPreKey.getKeyPair().getPublicKey().getPublicKey())

            signedPreKeyNode_sigNode = ProtocolTreeNode("signature",
                                                        data = signedPreKey.getSignature())

            signedPreKeyNode.addChildren([signedPreKeyNode_idNode, signedPreKeyNode_valueNode, signedPreKeyNode_sigNode])

            preKey = KeyHelper.generatePreKeys(i * 10, 1)[0]

            preKeyNode = ProtocolTreeNode("key")
            preKeyNode_idNode = ProtocolTreeNode("id", data = ResultGetKeysIqProtocolEntity._intToBytes(preKey.getId()))
            preKeyNode_valNode = ProtocolTreeNode("value", data = preKey.getKeyPair().getPublicKey().getPublicKey())
            preKeyNode.addChildren([preKeyNode_idNode, preKeyNode_valNode])

            userNode.addChildren([
                registrationNode,
                typeNode,
                identityNode,
                signedPreKeyNode,
                preKeyNode
            ])




from yowsup.layers import YowLayer, YowLayerEvent, YowProtocolLayer
from .protocolentities import *
from yowsup.layers.protocol_iq.protocolentities import ErrorIqProtocolEntity
class YowPresenceProtocolLayer(YowProtocolLayer):
    def __init__(self):
        handleMap = {
            "presence": (self.recvPresence, self.sendPresence),
            "iq":       (None, self.sendIq)
        }
        super(YowPresenceProtocolLayer, self).__init__(handleMap)

    def __str__(self):
        return "Presence Layer"

    def sendPresence(self, entity):
        self.entityToLower(entity)

    def recvPresence(self, node):
        self.toUpper(PresenceProtocolEntity.fromProtocolTreeNode(node))

    def sendIq(self, entity):
        if entity.getXmlns() == LastseenIqProtocolEntity.XMLNS:
            self._sendIq(entity, self.onLastSeenSuccess, self.onLastSeenError)

    def onLastSeenSuccess(self, protocolTreeNode, lastSeenEntity):
        self.toUpper(ResultLastseenIqProtocolEntity.fromProtocolTreeNode(protocolTreeNode))

    def onLastSeenError(self, protocolTreeNode, lastSeenEntity):
        self.toUpper(ErrorIqProtocolEntity.fromProtocolTreeNode(protocolTreeNode))



from .message_media import MediaMessageProtocolEntity
from yowsup.common.tools import WATools
from yowsup.common.tools import MimeTools
import os
class DownloadableMediaMessageProtocolEntity(MediaMessageProtocolEntity):
    '''
    <message t="{{TIME_STAMP}}" from="{{CONTACT_JID}}" 
        offline="{{OFFLINE}}" type="text" id="{{MESSAGE_ID}}" notify="{{NOTIFY_NAME}}">
        <media type="{{DOWNLOADABLE_MEDIA_TYPE: (image | audio | video)}}"
            mimetype="{{MIME_TYPE}}" 
            filehash="{{FILE_HASH}}"
            url="{{DOWNLOAD_URL}}" 
            ip="{{IP}}"
            size="{{MEDIA SIZE}}"
            file="{{FILENAME}}" 

            > {{THUMBNAIL_RAWDATA (JPEG?)}}
        </media>
    </message>
    '''
    def __init__(self, mediaType,
            mimeType, fileHash, url, ip, size, fileName, 
            _id = None, _from = None, to = None, notify = None, timestamp = None, 
            participant = None, preview = None, offline = None, retry = None):

        super(DownloadableMediaMessageProtocolEntity, self).__init__(mediaType, _id, _from, to, notify, timestamp, participant, preview, offline, retry)
        self.setDownloadableMediaProps(mimeType, fileHash, url, ip, size, fileName)

    def __str__(self):
        out  = super(DownloadableMediaMessageProtocolEntity, self).__str__()
        out += "MimeType: %s\n" % self.mimeType
        out += "File Hash: %s\n" % self.fileHash
        out += "URL: %s\n" % self.url
        out += "IP: %s\n" % self.ip
        out += "File Size: %s\n" % self.size
        out += "File name: %s\n" % self.fileName
        return out

    def getMediaSize(self):
        return self.size

    def getMediaUrl(self):
        return self.url

    def getMimeType(self):
        return self.mimeType

    def setDownloadableMediaProps(self, mimeType, fileHash, url, ip, size, fileName):
        self.mimeType   = mimeType
        self.fileHash   = fileHash
        self.url        = url
        self.ip         = ip
        self.size       = int(size)
        self.fileName   = fileName

    def toProtocolTreeNode(self):
        node = super(DownloadableMediaMessageProtocolEntity, self).toProtocolTreeNode()
        mediaNode = node.getChild("media")
        mediaNode.setAttribute("mimetype",  self.mimeType)
        mediaNode.setAttribute("filehash",  self.fileHash)
        mediaNode.setAttribute("url",       self.url)
        if self.ip:
            mediaNode.setAttribute("ip",        self.ip)
        mediaNode.setAttribute("size",      str(self.size))
        mediaNode.setAttribute("file",      self.fileName)

        return node

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = MediaMessageProtocolEntity.fromProtocolTreeNode(node)
        entity.__class__ = DownloadableMediaMessageProtocolEntity
        mediaNode = node.getChild("media")
        entity.setDownloadableMediaProps(
            mediaNode.getAttributeValue("mimetype"),
            mediaNode.getAttributeValue("filehash"),
            mediaNode.getAttributeValue("url"),
            mediaNode.getAttributeValue("ip"),
            mediaNode.getAttributeValue("size"),
            mediaNode.getAttributeValue("file")
            )
        return entity

    @staticmethod
    def fromFilePath(fpath, url, mediaType, ip, to, mimeType = None, preview = None, filehash = None, filesize = None):
        mimeType = mimeType or MimeTools.getMIME(fpath)
        filehash = filehash or WATools.getFileHashForUpload(fpath)
        size = filesize or os.path.getsize(fpath)
        fileName = os.path.basename(fpath)

        return DownloadableMediaMessageProtocolEntity(mediaType, mimeType, filehash, url, ip, size, fileName, to = to, preview = preview)





import json, sys
from xml.dom import minidom
import plistlib
import logging
logger = logging.getLogger(__name__)

class ResponseParser(object):
    def __init__(self):
        self.meta = "*"
        
    def parse(self, text, pvars):
        return text
    
    def getMeta(self):
        return self.meta
    
    
    def getVars(self, pvars):

        if type(pvars) is dict:
            return pvars

        if type(pvars) is list:
            
            out = {}
            
            for p in pvars:
                out[p] = p
                
            return out

class XMLResponseParser(ResponseParser):
    
    def __init__(self):
        
        try:
            import libxml2
        except ImportError:
            print("libxml2 XMLResponseParser requires libxml2")
            sys.exit(1)

        self.meta = "text/xml";

    def parse(self, xml, pvars):
        import libxml2
        doc = libxml2.parseDoc(xml)
        
        pvars = self.getVars(pvars)
        vals = {}
        for k, v in pvars.items():
            res = doc.xpathEval(v)
            vals[k] = []
            for r in res:
                
                #if not vals.has_key(r.name):
                #   vals[r.name] = []
                
                if r.type == 'element':
                    #vals[r.name].append(self.xmlToDict(minidom.parseString(str(r)))[r.name])
                    vals[k].append(self.xmlToDict(minidom.parseString(str(r)))[r.name])
                elif r.type == 'attribute':
                    vals[k].append(r.content)
                else:
                    logger.error("UNKNOWN TYPE")
            
            if len(vals[k]) == 1:
                vals[k] = vals[k][0]
            elif len(vals[k]) == 0:
                vals[k] = None

        return vals
    
    def xmlToDict(self, xmlNode):
        if xmlNode.nodeName == "#document":
            
            node = {xmlNode.firstChild.nodeName:{}}
            
            node[xmlNode.firstChild.nodeName] = self.xmlToDict(xmlNode.firstChild)
            return node
        
        node = {}
        curr = node
        
        if xmlNode.attributes:
            for name, value in xmlNode.attributes.items():
                curr[name] = value

        for n in xmlNode.childNodes:
            
            if n.nodeType == n.TEXT_NODE:
                curr["__TEXT__"] = n.data
                continue
            
            if not n.nodeName in curr:
                curr[n.nodeName] = []

            if len(xmlNode.getElementsByTagName(n.nodeName)) > 1:
                #curr[n.nodeName] = []
                curr[n.nodeName].append(self.xmlToDict(n))
            else:
                curr[n.nodeName] = self.xmlToDict(n)
            
            
        return node

class JSONResponseParser(ResponseParser):
    
    def __init__(self):
        self.meta = "text/json"

    def parse(self, jsonData, pvars):
        
        d = json.loads(jsonData)
        pvars = self.getVars(pvars)
        
        parsed = {}     
        
        for k,v in pvars.items():
            parsed[k] = self.query(d, v)

        return parsed
    
    def query(self, d, key):
        keys = key.split('.', 1)
            
        currKey = keys[0]
        
        if(currKey in d):
            item = d[currKey]
            
            if len(keys) == 1:
                    return item
            
            if type(item) is dict:
                return self.query(item, keys[1])
            
            elif type(item) is list:
                output = []

                for i in item:
                    output.append(self.query(i, keys[1]))
                return output
            
            else:
                return None

class PListResponseParser(ResponseParser):
    def __init__(self):
        self.meta = "text/xml"
    
    def parse(self, xml, pvars):
        
        #tmp = minidom.parseString(xml)
        
        if sys.version_info >= (3, 0):
            pl = plistlib.readPlistFromBytes(xml.encode());
        else:
            pl = plistlib.readPlistFromString(xml);
        
        parsed= {}
        pvars = self.getVars(pvars)
        
        for k,v in pvars.items():
            parsed[k] = pl[k] if  k in pl else None
        
        return parsed;
        



from yowsup.layers.protocol_contacts.protocolentities.iq_sync_get import GetSyncIqProtocolEntity
from yowsup.structs.protocolentity import ProtocolEntityTest
import unittest

entity = GetSyncIqProtocolEntity(["12345678", "8764543121"])

class GetSyncIqProtocolEntityTest(ProtocolEntityTest, unittest.TestCase):
    def setUp(self):
        self.ProtocolEntity = GetSyncIqProtocolEntity
        self.node = entity.toProtocolTreeNode()


from yowsup.structs import ProtocolEntity, ProtocolTreeNode
from .ack import AckProtocolEntity
class OutgoingAckProtocolEntity(AckProtocolEntity):

    '''
    <ack type="{{delivery | read}}" class="{{message | receipt | ?}}" id="{{MESSAGE_ID}} to={{TO_JID}}">
    </ack>

    <ack to="{{GROUP_JID}}" participant="{{JID}}" id="{{MESSAGE_ID}}" class="receipt" type="{{read | }}">
    </ack>

    '''

    def __init__(self, _id, _class, _type, _to, _participant = None):
        super(OutgoingAckProtocolEntity, self).__init__(_id, _class)
        self.setOutgoingData(_type, _to, _participant)

    def setOutgoingData(self, _type, _to, _participant):
        self._type = _type
        self._to = _to
        self._participant = _participant
    
    def toProtocolTreeNode(self):
        node = super(OutgoingAckProtocolEntity, self).toProtocolTreeNode()
        if self._type:
            node.setAttribute("type", self._type)
        node.setAttribute("to", self._to)
        if self._participant:
            node.setAttribute("participant", self._participant)
        return node

    def __str__(self):
        out  = super(OutgoingAckProtocolEntity, self).__str__()
        out += "Type: %s\n" % self._type
        out += "To: %s\n" % self._to
        if self._participant:
            out += "Participant: %s\n" % self._participant
        return out

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = AckProtocolEntity.fromProtocolTreeNode(node)
        entity.__class__ = OutgoingAckProtocolEntity
        entity.setOutgoingData(
            node.getAttributeValue("type"),
            node.getAttributeValue("to"),
            node.getAttributeValue("participant")
        )
        return entity



class YowConstants:
    DOMAIN       = "s.whatsapp.net"
    ENDPOINTS     = (
        ("e1.whatsapp.net", 443),
        ("e2.whatsapp.net", 443),
        ("e3.whatsapp.net", 443),
        ("e4.whatsapp.net", 443),
        ("e5.whatsapp.net", 443),
        ("e6.whatsapp.net", 443),
        ("e7.whatsapp.net", 443),
        ("e8.whatsapp.net", 443),
        ("e9.whatsapp.net", 443),
        ("e10.whatsapp.net", 443),
        ("e11.whatsapp.net", 443),
        ("e12.whatsapp.net", 443),
        ("e13.whatsapp.net", 443),
        ("e14.whatsapp.net", 443),
        ("e15.whatsapp.net", 443),
        ("e16.whatsapp.net", 443),
        )

    WHATSAPP_SERVER = "s.whatsapp.net"
    WHATSAPP_GROUP_SERVER = "g.us"

    PATH_STORAGE = "~/.yowsup"

    PREVIEW_WIDTH = 64
    PREVIEW_HEIGHT = 64



from yowsup.structs import ProtocolTreeNode
import math
class ReadDecoder:
    def __init__(self, tokenDictionary):
        self.streamStarted = False;
        self.tokenDictionary = tokenDictionary

    def reset(self):
        self.streamStarted = False

    def getProtocolTreeNode(self, data):
        if not self.streamStarted:
            return self.streamStart(data)
        return self.nextTreeInternal(data)

    def _getToken(self, index, data):
        token = self.tokenDictionary.getToken(index)
        if not token:
            index = self.readInt8(data)
            token = self.tokenDictionary.getToken(index, True)
            if not token:
                raise ValueError("Invalid token %s" % token)

        return token

    def streamStart(self, data):
        self.streamStarted = True
        tag = data.pop(0)
        size = self.readListSize(tag, data)
        tag = data.pop(0)

        if tag != 1:
            if tag == 236:
                tag = data.pop(0) + 237
            token = self._getToken(tag, data)#self.tokenDictionary.getToken(tag)
            raise Exception("expecting STREAM_START in streamStart, instead got token: %s" % token)
        attribCount = (size - 2 + size % 2) / 2
        self.readAttributes(attribCount, data)

    def readNibble(self, data):
        _byte = self.readInt8(data)
        ignoreLastNibble = bool(_byte & 0x80)
        size = (_byte & 0x7f);
        nrOfNibbles = size * 2 - int(ignoreLastNibble)
        dataArr = self.readArray(size, data)
        string = ''
        for i in range(0, nrOfNibbles):
            _byte = dataArr[int(math.floor(i/2))]
            _shift = 4 * (1 - i % 2)
            dec = (_byte & (15 << _shift)) >> _shift

            if dec in (0,1,2,3,4,5,6,7,8,9):
                string += str(dec)
            elif dec in (10, 11):
                string += chr(dec - 10 + 45)
            else:
                raise Exception("Bad nibble %s" % dec)
        return string





    def readInt8(self,data):
        return data.pop(0);

    def readInt16(self, data):
        intTop = data.pop(0)
        intBot = data.pop(0)
        value = (intTop << 8) + intBot
        if value is not None:
            return value
        else:
            return ""

    def readInt24(self,data):
        int1 = data.pop(0)
        int2 = data.pop(0)
        int3 = data.pop(0)
        value = (int1 << 16) + (int2 << 8) + (int3 << 0)
        return value


    def readListSize(self,token, data):
        size = 0
        if token == 0:
            size = 0
        else:
            if token == 248:
                size = self.readInt8(data)
            else:
                if token == 249:
                    size = self.readInt16(data)
                else:
                    raise Exception("invalid list size in readListSize: token " + str(token))
        return size

    def readAttributes(self, attribCount, data):
        attribs = {}

        for i in range(0, int(attribCount)):
            key = self.readString(data.pop(0), data)
            value = self.readString(data.pop(0), data)
            attribs[key]=value
        return attribs


    def readString(self,token, data):

        if token == -1:
            raise Exception("-1 token in readString")

        if token > 2 and token < 245:
            return self._getToken(token, data)

        if token == 0:
            return None


        if token == 252:
            size8 = self.readInt8(data)
            buf8 = self.readArray(size8, data)
            return "".join(map(chr, buf8))


        if token == 253:
            size24 = self.readInt24(data)
            buf24 = self.readArray(size24, data)
            return "".join(map(chr, buf24))

        if token == 250:
            user = self.readString(data.pop(0), data)
            server = self.readString(data.pop(0), data)
            if user is not None and server is not None:
                return user + "@" + server
            if server is not None:
                return server
            raise Exception("readString couldn't reconstruct jid")
        elif token == 255:
            return self.readNibble(data)

        raise Exception("readString couldn't match token "+str(token))

    def readArray(self, length, data):
        out = []
        for i in range(0, length):
            out.append(data.pop(0))

        return out

    def nextTreeInternal(self, data):
        b = data.pop(0)

        size = self.readListSize(b, data)
        b = data.pop(0)
        if b == 2:
            return None


        tag = self.readString(b, data)
        if size == 0 or tag is None:
            raise ValueError("nextTree sees 0 list or null tag")

        attribCount = (size - 2 + size % 2)/2;
        attribs = self.readAttributes(attribCount, data)
        if size % 2 ==1:
            return ProtocolTreeNode(tag, attribs)

        b = data.pop(0)

        if self.isListTag(b):
            return ProtocolTreeNode(tag,attribs,self.readList(b, data))

        return ProtocolTreeNode(tag, attribs, None, self.readString(b, data))

    def readList(self,token, data):
        size = self.readListSize(token, data)
        listx = []
        for i in range(0,size):
            listx.append(self.nextTreeInternal(data))

        return listx;

    def isListTag(self, b):
        return b in (248, 0, 249)



from yowsup.layers.protocol_iq.protocolentities.test_iq import IqProtocolEntityTest
from yowsup.layers.protocol_media.protocolentities import RequestUploadIqProtocolEntity
from yowsup.structs import ProtocolTreeNode
class RequestUploadIqProtocolEntityTest(IqProtocolEntityTest):
    def setUp(self):
        super(RequestUploadIqProtocolEntityTest, self).setUp()
        mediaNode = ProtocolTreeNode("media", {"hash": "hash", "size": "1234", "orighash": "orighash", "type": "image"})
        self.ProtocolEntity = RequestUploadIqProtocolEntity
        self.node.setAttribute("type", "set")
        self.node.addChild(mediaNode)


from yowsup.layers.protocol_media.protocolentities.test_message_media import MediaMessageProtocolEntityTest
from yowsup.layers.protocol_media.protocolentities import VCardMediaMessageProtocolEntity
from yowsup.structs import ProtocolTreeNode
class VCardMediaMessageProtocolEntityTest(MediaMessageProtocolEntityTest):
    def setUp(self):
        super(VCardMediaMessageProtocolEntityTest, self).setUp()
        self.ProtocolEntity = VCardMediaMessageProtocolEntity
        vcardNode = ProtocolTreeNode("vcard", {"name":"abc"}, None, "VCARD_DATA")
        mediaNode = self.node.getChild("media")
        mediaNode["type"] = "vcard"
        mediaNode.addChild(vcardNode)



from yowsup.common import YowConstants
from yowsup.layers.protocol_iq.protocolentities import ResultIqProtocolEntity
from yowsup.structs import ProtocolTreeNode
from axolotl.state.prekeybundle import PreKeyBundle
from axolotl.identitykey import IdentityKey
from axolotl.ecc.curve import Curve
from axolotl.ecc.djbec import DjbECPublicKey
import binascii
import sys
class ResultGetKeysIqProtocolEntity(ResultIqProtocolEntity):
    """
    <iq type="result" from="s.whatsapp.net" id="3">
    <list>
    <user jid="79049347231@s.whatsapp.net">
    <registration>
        HEX:7a9cec4b</registration>
    <type>

    HEX:05</type>
    <identity>
    HEX:eeb668c8d062c99b43560c811acfe6e492798b496767eb060d99e011d3862369</identity>
    <skey>
    <id>

    HEX:000000</id>
    <value>
    HEX:a1b5216ce4678143fb20aaaa2711a8c2b647230164b79414f0550b4e611ccd6c</value>
    <signature>
    HEX:94c231327fcd664b34603838b5e9ba926718d71c206e92b2b400f5cf4ae7bf17d83557bf328c1be6d51efdbd731a26d000adb8f38f140b1ea2a5fd3df2688085</signature>
        </skey>
    <key>
        <id>
        HEX:36b545</id>
    <value>
    HEX:c20826f622bec24b349ced38f1854bdec89ba098ef4c06b2402800d33e9aff61</value>
    </key>
    </user>
    </list>
    </iq>
    """
    def __init__(self, _id, preKeyBundleMap = None):
        super(ResultGetKeysIqProtocolEntity, self).__init__(_from = YowConstants.WHATSAPP_SERVER, _id=_id)
        self.setPreKeyBundleMap(preKeyBundleMap)

    def getJids(self):
        return self.preKeyBundleMap.keys()

    def setPreKeyBundleMap(self, preKeyBundleMap = None):
        self.preKeyBundleMap = preKeyBundleMap or {}

    def setPreKeyBundleFor(self, jid, preKeyBundle):
        self.preKeyBundleMap[jid] = preKeyBundle

    def getPreKeyBundleFor(self, jid):
        if jid in self.preKeyBundleMap:
            return self.preKeyBundleMap[jid]

    @staticmethod
    def _intToBytes(val):
        return binascii.unhexlify(format(val, 'x').zfill(8).encode())

    @staticmethod
    def _bytesToInt(val):
        if sys.version_info >= (3,0):
            valEnc = val.encode('latin-1') if type(val) is str else val
        else:
            valEnc = val
        return int(binascii.hexlify(valEnc), 16)

    @staticmethod
    def encStr(string):
        if sys.version_info >= (3,0) and type(string) is str:
            return string.encode('latin-1')
        return string

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = ResultIqProtocolEntity.fromProtocolTreeNode(node)
        entity.__class__ = ResultGetKeysIqProtocolEntity
        entity.setPreKeyBundleMap()
        userNodes = node.getChild("list").getAllChildren()
        for userNode in userNodes:
            preKeyNode = userNode.getChild("key")
            signedPreKeyNode = userNode.getChild("skey")
            registrationId = ResultGetKeysIqProtocolEntity._bytesToInt(userNode.getChild("registration").getData())
            identityKey = IdentityKey(DjbECPublicKey(ResultGetKeysIqProtocolEntity.encStr(userNode.getChild("identity").getData())))

            preKeyId = ResultGetKeysIqProtocolEntity._bytesToInt(preKeyNode.getChild("id").getData())
            preKeyPublic = DjbECPublicKey(ResultGetKeysIqProtocolEntity.encStr(preKeyNode.getChild("value").getData()))

            signedPreKeyId = ResultGetKeysIqProtocolEntity._bytesToInt(signedPreKeyNode.getChild("id").getData())
            signedPreKeySig = ResultGetKeysIqProtocolEntity.encStr(signedPreKeyNode.getChild("signature").getData())
            signedPreKeyPub = DjbECPublicKey(ResultGetKeysIqProtocolEntity.encStr(signedPreKeyNode.getChild("value").getData()))

            preKeyBundle = PreKeyBundle(registrationId, 1, preKeyId, preKeyPublic,
                                        signedPreKeyId, signedPreKeyPub, signedPreKeySig, identityKey)

            entity.setPreKeyBundleFor(userNode["jid"], preKeyBundle)

        return entity


    def toProtocolTreeNode(self):
        node = super(ResultGetKeysIqProtocolEntity, self).toProtocolTreeNode()
        listNode = ProtocolTreeNode("list")
        node.addChild(listNode)

        for jid, preKeyBundle in self.preKeyBundleMap.items():
            userNode = ProtocolTreeNode("user", {"jid": jid})
            registrationNode = ProtocolTreeNode("registration", data = self.__class__._intToBytes(preKeyBundle.getRegistrationId()))
            typeNode = ProtocolTreeNode("type", data = self.__class__._intToBytes(Curve.DJB_TYPE))
            identityNode = ProtocolTreeNode("identity", data = preKeyBundle.getIdentityKey().getPublicKey().getPublicKey())

            skeyNode = ProtocolTreeNode("skey")
            skeyNode_idNode = ProtocolTreeNode("id", data=self.__class__._intToBytes(preKeyBundle.getSignedPreKeyId()))
            skeyNode_valueNode = ProtocolTreeNode("value", data=preKeyBundle.getSignedPreKey().getPublicKey())
            skeyNode_signatureNode = ProtocolTreeNode("signature", data=preKeyBundle.getSignedPreKeySignature())
            skeyNode.addChildren([skeyNode_idNode, skeyNode_valueNode, skeyNode_signatureNode])

            preKeyNode = ProtocolTreeNode("key")
            preKeyNode_idNode = ProtocolTreeNode("id", data = self.__class__._intToBytes(preKeyBundle.getPreKeyId()))
            preKeyNode_valueNode = ProtocolTreeNode("value", data= preKeyBundle.getPreKey().getPublicKey())
            preKeyNode.addChildren([preKeyNode_idNode, preKeyNode_valueNode])

            userNode.addChildren([
                registrationNode,
                typeNode,
                identityNode,
                skeyNode,
                preKeyNode
            ])
            listNode.addChild(userNode)

        return node




















from yowsup.structs import ProtocolEntity, ProtocolTreeNode
class ReceiptProtocolEntity(ProtocolEntity):

    '''
    delivered:
    <receipt to="xxxxxxxxxxx@s.whatsapp.net" id="1415389947-15"></receipt>

    read
    <receipt to="xxxxxxxxxxx@s.whatsapp.net" id="1415389947-15" type="read || played"></receipt>

    INCOMING
    <receipt offline="0" from="4915225256022@s.whatsapp.net" id="1415577964-1" t="1415578027" type="played?"></receipt>
    '''

    def __init__(self, _id):
        super(ReceiptProtocolEntity, self).__init__("receipt")
        self._id = _id

    def getId(self):
        return self._id
    
    def toProtocolTreeNode(self):
        attribs = {
            "id"           : self._id
        }
        return self._createProtocolTreeNode(attribs, None, data = None)


    def __str__(self):
        out  = "Receipt:\n"
        out += "ID: %s\n" % self._id
        return out

    @staticmethod
    def fromProtocolTreeNode(node):
        return ReceiptProtocolEntity(
            node.getAttributeValue("id")
            )



from yowsup.common import YowConstants
from yowsup.layers.protocol_iq.protocolentities import IqProtocolEntity
from yowsup.structs import ProtocolTreeNode
import hashlib
import base64
import os
from yowsup.common.tools import WATools
class RequestUploadIqProtocolEntity(IqProtocolEntity):
    '''
    <iq to="s.whatsapp.net" type="set" xmlns="w:m">
        <media hash="{{b64_hash}}" type="{{type}}" size="{{size_bytes}}" orighash={{b64_orighash?}}></media>
    </iq>
    '''

    MEDIA_TYPE_IMAGE = "image"
    MEDIA_TYPE_VIDEO = "video"
    MEDIA_TYPE_AUDIO = "audio"
    XMLNS = "w:m"

    TYPES_MEDIA = (MEDIA_TYPE_AUDIO, MEDIA_TYPE_IMAGE, MEDIA_TYPE_VIDEO)

    def __init__(self, mediaType, b64Hash = None, size = None, origHash = None, filePath = None ):
        super(RequestUploadIqProtocolEntity, self).__init__("w:m", _type = "set", to = YowConstants.WHATSAPP_SERVER)

        assert (b64Hash and size) or filePath, "Either specify hash and size, or specify filepath and let me generate the rest"

        if filePath:
            assert os.path.exists(filePath), "Either specified path does not exist, or yowsup doesn't have permission to read: %s" % filePath
            b64Hash = self.__class__.getFileHashForUpload(filePath)

            size = os.path.getsize(filePath)


        self.setRequestArguments(mediaType, b64Hash, size, origHash)

    def setRequestArguments(self, mediaType, b64Hash, size, origHash = None):
        assert mediaType in self.__class__.TYPES_MEDIA, "Expected media type to be in %s, got %s" % (self.__class__.TYPES_MEDIA, mediaType)
        self.mediaType = mediaType
        self.b64Hash = b64Hash
        self.size = int(size)
        self.origHash = origHash

    @staticmethod
    def getFileHashForUpload(filePath):
        return WATools.getFileHashForUpload(filePath)

    def __str__(self):
        out = super(RequestUploadIqProtocolEntity, self).__str__()
        out += "Media Type: %s\n" % self.mediaType
        out += "B64Hash: %s\n" % self.b64Hash
        out += "Size: %s\n" % self.size
        if self.origHash:
            out += "OrigHash: %s\n" % self.origHash
        return out

    def toProtocolTreeNode(self):
        node = super(RequestUploadIqProtocolEntity, self).toProtocolTreeNode()
        attribs = {
            "hash": self.b64Hash,
            "type": self.mediaType,
            "size": str(self.size)
        }
        if self.origHash:
            attribs["orighash"] = self.origHash
        mediaNode = ProtocolTreeNode("media", attribs)
        node.addChild(mediaNode)
        return node

    @staticmethod
    def fromProtocolTreeNode(node):
        assert node.getAttributeValue("type") == "set", "Expected set as iq type in request upload, got %s" % node.getAttributeValue("type")
        entity = IqProtocolEntity.fromProtocolTreeNode(node)
        entity.__class__ = RequestUploadIqProtocolEntity
        mediaNode = node.getChild("media")
        entity.setRequestArguments(
            mediaNode.getAttributeValue("type"),
            mediaNode.getAttributeValue("hash"),
            mediaNode.getAttributeValue("size"),
            mediaNode.getAttributeValue("orighash")
        )
        return entity



from yowsup.structs import ProtocolEntity, ProtocolTreeNode
from .ib import IbProtocolEntity
class DirtyIbProtocolEntity(IbProtocolEntity):
    '''
    <ib>
        <dirty type="{{groups | ?}}" timestamp="{{ts}}"></dirty>
    </ib>
    '''
    def __init__(self, timestamp, _type):
        super(DirtyIbProtocolEntity, self).__init__()
        self.setProps(timestamp, _type)


    def setProps(self, timestamp, _type):
        self.timestamp = int(timestamp)
        self._type = _type
    
    def toProtocolTreeNode(self):
        node = super(DirtyIbProtocolEntity, self).toProtocolTreeNode()
        dirtyNode = ProtocolTreeNode("dirty")
        dirtyNode["timestamp"] = str(self.timestamp)
        dirtyNode["type"] = self._type
        node.addChild(dirtyNode)
        return node

    def __str__(self):
        out = super(DirtyIbProtocolEntity, self).__str__()
        out += "Type: %s\n" % self._type
        out += "Timestamp: %s\n" % self.timestamp
        return out

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = IbProtocolEntity.fromProtocolTreeNode(node)
        entity.__class__ = DirtyIbProtocolEntity
        dirtyChild = node.getChild("dirty")
        entity.setProps(dirtyChild["timestamp"], dirtyChild["type"])
        return entity



from yowsup.layers.auth import YowAuthenticationProtocolLayer
from yowsup.layers import YowLayerEvent, YowLayerTest
from yowsup.layers.auth.protocolentities import StreamFeaturesProtocolEntity,AuthProtocolEntity, ChallengeProtocolEntity
from yowsup.structs import ProtocolTreeNode
from yowsup.stacks import YowStack
import base64
# from unittest.mock import MagicMock


class AuthenticationProtocolLayerTest(YowLayerTest, YowAuthenticationProtocolLayer):
    def setUp(self):
        #YowAuthenticatorLayer.__init__(self)
        super(YowAuthenticationProtocolLayer, self).__init__()
        dummyStack = YowStack()
        self.setStack(dummyStack)

        self.credentials =  ("dummyusername", bytearray("password", "latin-1"))
        dummyStack.setProp(YowAuthenticationProtocolLayer.PROP_CREDENTIALS, self.credentials)
        #ticatorLayer.setProp("credentials", self.credentials)

    def test_streamfeatures(self):
        self._sendFeatures()
        self.assertEqual(self.lowerSink.pop(), StreamFeaturesProtocolEntity(["readreceipts","groups_v2","privacy","presence"]).toProtocolTreeNode())

    def test_auth(self):
        self._sendAuth()
        self.assertEqual(self.lowerSink.pop(), AuthProtocolEntity(self.credentials[0]).toProtocolTreeNode())

    def test_handle_challenge(self):
        node = ProtocolTreeNode("challenge", {}, None, "salt")
        self.handleChallenge(node)



    def test_login_onconnected(self):
        self.onEvent(YowLayerEvent("network.state.connected"))



from yowsup.structs import ProtocolEntity, ProtocolTreeNode
from .message_media_downloadable import DownloadableMediaMessageProtocolEntity
from yowsup.common.tools import ImageTools
class ImageDownloadableMediaMessageProtocolEntity(DownloadableMediaMessageProtocolEntity):
    '''
    <message t="{{TIME_STAMP}}" from="{{CONTACT_JID}}" 
        offline="{{OFFLINE}}" type="text" id="{{MESSAGE_ID}}" notify="{{NOTIFY_NAME}}">
        <media type="{{DOWNLOADABLE_MEDIA_TYPE: (image | audio | video)}}"
            mimetype="{{MIME_TYPE}}" 
            filehash="{{FILE_HASH}}"
            url="{{DOWNLOAD_URL}}" 
            ip="{{IP}}"
            size="{{MEDIA SIZE}}"
            file="{{FILENAME}}" 


            encoding="{{ENCODING}}" 
            height="{{IMAGE_HEIGHT}}" 
            width="{{IMAGE_WIDTH}}"
            
            > {{THUMBNAIL_RAWDATA (JPEG?)}}
        </media>
    </message>
    '''
    def __init__(self,
            mimeType, fileHash, url, ip, size, fileName,
            encoding, width, height, caption = None, 
            _id = None, _from = None, to = None, notify = None, timestamp = None, 
            participant = None, preview = None, offline = None, retry = None):

        super(ImageDownloadableMediaMessageProtocolEntity, self).__init__("image",
            mimeType, fileHash, url, ip, size, fileName,
            _id, _from, to, notify, timestamp, participant, preview, offline, retry)
        self.setImageProps(encoding, width, height, caption)

    def __str__(self):
        out  = super(ImageDownloadableMediaMessageProtocolEntity, self).__str__()
        out += "Encoding: %s\n" % self.encoding
        out += "Width: %s\n" % self.width
        out += "Height: %s\n" % self.height
        if self.caption:
            out += "Caption: %s\n" % self.caption
        return out

    def setImageProps(self, encoding, width, height, caption):
        self.encoding   = encoding
        self.width      = int(width)
        self.height     = int(height)
        self.caption    = caption

    def getCaption(self):
        return self.caption
        
    def toProtocolTreeNode(self):
        node = super(ImageDownloadableMediaMessageProtocolEntity, self).toProtocolTreeNode()
        mediaNode = node.getChild("media")

        mediaNode.setAttribute("encoding",  self.encoding)
        mediaNode.setAttribute("width",     str(self.width))
        mediaNode.setAttribute("height",    str(self.height))
        if self.caption:
            mediaNode.setAttribute("caption", self.caption)

        return node

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = DownloadableMediaMessageProtocolEntity.fromProtocolTreeNode(node)
        entity.__class__ = ImageDownloadableMediaMessageProtocolEntity
        mediaNode = node.getChild("media")
        entity.setImageProps(
            mediaNode.getAttributeValue("encoding"),
            mediaNode.getAttributeValue("width"),
            mediaNode.getAttributeValue("height"),
            mediaNode.getAttributeValue("caption"),
        )
        return entity


    @staticmethod
    def fromFilePath(path, url, ip, to, mimeType = None, caption = None, dimensions = None):
        preview = ImageTools.generatePreviewFromImage(path)
        entity = DownloadableMediaMessageProtocolEntity.fromFilePath(path, url, DownloadableMediaMessageProtocolEntity.MEDIA_TYPE_IMAGE, ip, to, mimeType, preview)
        entity.__class__ = ImageDownloadableMediaMessageProtocolEntity

        if not dimensions:
            dimensions = ImageTools.getImageDimensions(path)

        assert dimensions, "Could not determine image dimensions"

        width, height = dimensions
        entity.setImageProps("raw", width, height, caption)
        return entity




from .iq_picture import PictureIqProtocolEntity
from yowsup.structs import ProtocolTreeNode
class ResultGetPictureIqProtocolEntity(PictureIqProtocolEntity):
    '''
    <iq type="result" from="{{jid}}" id="{{id}}">
        <picture type="image | preview" id="{{another_id}}">
        {{Binary bytes of the picture.}}
        </picture>
    </iq>
    '''
    def __init__(self, jid, pictureData, pictureId, preview = True, _id = None):
        super(ResultGetPictureIqProtocolEntity, self).__init__(jid, _id, "result")
        self.setResultPictureProps(pictureData, pictureId, preview)

    def setResultPictureProps(self, pictureData, pictureId, preview = True):
        self.preview = preview
        self.pictureData = pictureData
        self.pictureId = pictureId

    def isPreview(self):
        return self.preview

    def getPictureData(self):
        return self.pictureData

    def getPictureId(self):
        return self.pictureId

    def writeToFile(self, path):
        with open(path, "wb") as outFile:
            outFile.write(self.getPictureData())

    def toProtocolTreeNode(self):
        node = super(ResultGetPictureIqProtocolEntity, self).toProtocolTreeNode()
        pictureNode = ProtocolTreeNode({"type": "preview" if self.isPreview() else "image" }, data = self.getPictureData())
        node.addChild(pictureNode)
        return node

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = PictureIqProtocolEntity.fromProtocolTreeNode(node)
        entity.__class__ = ResultGetPictureIqProtocolEntity
        pictureNode = node.getChild("picture")
        entity.setResultPictureProps(pictureNode.getData(), pictureNode.getAttributeValue("id"), pictureNode.getAttributeValue("type") == "preview")
        return entity


from yowsup.layers.protocol_iq.protocolentities.iq import IqProtocolEntity
from yowsup.structs.protocoltreenode import ProtocolTreeNode
class LastseenIqProtocolEntity(IqProtocolEntity):
    XMLNS = "jabber:iq:last"
    def __init__(self, jid, _id = None):
        super(LastseenIqProtocolEntity, self).__init__(self.__class__.XMLNS, _type = "get", to = jid, _id = _id)

    @staticmethod
    def fromProtocolTreeNode(node):
        return LastseenIqProtocolEntity(node["to"])

    def toProtocolTreeNode(self):
        node = super(LastseenIqProtocolEntity, self).toProtocolTreeNode()
        node.setAttribute("xmlns", self.__class__.XMLNS)
        node.addChild(ProtocolTreeNode("query"))
        return node



from yowsup.layers.protocol_iq.protocolentities import ErrorIqProtocolEntity
class FailureAddParticipantsIqProtocolEntity(ErrorIqProtocolEntity):
    '''
    <iq type="error" from="{{group_jid}}" id="{{id}}">
        <error text="item-not-found" code="404">
        </error>
    </iq>
    '''

    def __init__(self, _id, _from, _code, _text, _backoff= 0 ):
        super(FailureAddParticipantsIqProtocolEntity, self).__init__(_from = _from,
                                                                     _id = _id, code = _code,
                                                                     text = _text, backoff = _backoff)
    @staticmethod
    def fromProtocolTreeNode(node):
        entity = ErrorIqProtocolEntity.fromProtocolTreeNode(node)
        entity.__class__ = FailureAddParticipantsIqProtocolEntity
        return entity


from yowsup.layers.protocol_notifications.protocolentities.notification import NotificationProtocolEntity
from yowsup.structs import ProtocolTreeNode
from yowsup.structs.protocolentity import ProtocolEntityTest
import unittest

class NotificationProtocolEntityTest(ProtocolEntityTest, unittest.TestCase):
    def setUp(self):
        self.ProtocolEntity = NotificationProtocolEntity
        attribs = {
            "t": "12345",
            "from": "from_jid",
            "offline": "0",
            "type": "notif_type",
            "id": "message-id",
            "notify": "notify_name"
        }
        self.node = ProtocolTreeNode("notification", attribs)


from yowsup.layers.protocol_calls.protocolentities.call import CallProtocolEntity
from yowsup.structs import ProtocolTreeNode
from yowsup.structs.protocolentity import ProtocolEntityTest
import unittest

class CallProtocolEntityTest(ProtocolEntityTest, unittest.TestCase):
    def setUp(self):
        self.ProtocolEntity = CallProtocolEntity
        children = [ProtocolTreeNode("offer", {"call-id": "call_id"})]
        attribs = {
            "t": "12345",
            "from": "from_jid",
            "offline": "0",
            "id": "message_id",
            "notify": "notify_name"
        }
        self.node = ProtocolTreeNode("call", attribs, children)



from yowsup.structs import ProtocolEntity, ProtocolTreeNode
from .presence import PresenceProtocolEntity
class UnsubscribePresenceProtocolEntity(PresenceProtocolEntity):

    '''
    <presence type="unsubscribe" to="jid"></presence>
    '''

    def __init__(self, jid):
        super(UnsubscribePresenceProtocolEntity, self).__init__("unsubscribe")
        self.setProps(jid)

    def setProps(self, jid):
        self.jid = jid
    
    def toProtocolTreeNode(self):
        node = super(UnsubscribePresenceProtocolEntity, self).toProtocolTreeNode()
        node.setAttribute("to", self.jid)
        return node

    def __str__(self):
        out  = super(UnsubscribePresenceProtocolEntity, self).__str__()
        out += "To: %s\n" % self.jid
        return out

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = PresenceProtocolEntity.fromProtocolTreeNode(node)
        entity.__class__ = UnsubscribePresenceProtocolEntity
        entity.setProps(
            node.getAttributeValue("to")
        )
        return entity



from yowsup.layers.protocol_media.protocolentities.message_media_downloadable_image import ImageDownloadableMediaMessageProtocolEntity
from yowsup.layers.protocol_media.protocolentities.test_message_media_downloadable import DownloadableMediaMessageProtocolEntityTest

class ImageDownloadableMediaMessageProtocolEntityTest(DownloadableMediaMessageProtocolEntityTest):
    def setUp(self):
        super(ImageDownloadableMediaMessageProtocolEntityTest, self).setUp()
        self.ProtocolEntity = ImageDownloadableMediaMessageProtocolEntity
        mediaNode = self.node.getChild("media")
        mediaNode.setAttribute("encoding",  "ENCODING")
        mediaNode.setAttribute("width",     "1024")
        mediaNode.setAttribute("height",     "768")
        mediaNode.setAttribute("caption", "caption")



from yowsup.layers.protocol_media.protocolentities.message_media_downloadable_audio import AudioDownloadableMediaMessageProtocolEntity
from yowsup.layers.protocol_media.protocolentities.test_message_media_downloadable import DownloadableMediaMessageProtocolEntityTest

class AudioDownloadableMediaMessageProtocolEntityTest(DownloadableMediaMessageProtocolEntityTest):
    def setUp(self):
        super(AudioDownloadableMediaMessageProtocolEntityTest, self).setUp()
        self.ProtocolEntity = AudioDownloadableMediaMessageProtocolEntity
        mediaNode = self.node.getChild("media")
        mediaNode.setAttribute("abitrate", "31")
        mediaNode.setAttribute("acodec", "aac")
        mediaNode.setAttribute("asampfreq", "22050")
        mediaNode.setAttribute("duration", "3")
        mediaNode.setAttribute("encoding", "raw")
        mediaNode.setAttribute("origin", "live")
        mediaNode.setAttribute("seconds", "3")



from yowsup.layers.interface                           import YowInterfaceLayer, ProtocolEntityCallback
from yowsup.layers.protocol_contacts.protocolentities  import GetSyncIqProtocolEntity, ResultSyncIqProtocolEntity
from yowsup.layers.protocol_iq.protocolentities import ErrorIqProtocolEntity
import threading
import logging
logger = logging.getLogger(__name__)

class SyncLayer(YowInterfaceLayer):

    PROP_CONTACTS = "org.openwhatsapp.yowsup.prop.syncdemo.contacts"

    def __init__(self):
        super(SyncLayer, self).__init__()

    #call back function when there is a successful connection to whatsapp server
    @ProtocolEntityCallback("success")
    def onSuccess(self, successProtocolEntity):
        contacts= self.getProp(self.__class__.PROP_CONTACTS, [])
        contactEntity = GetSyncIqProtocolEntity(contacts)
        self._sendIq(contactEntity, self.onGetSyncResult, self.onGetSyncError)

    def onGetSyncResult(self, resultSyncIqProtocolEntity, originalIqProtocolEntity):
        print(resultSyncIqProtocolEntity)
        raise KeyboardInterrupt()

    def onGetSyncError(self, errorSyncIqProtocolEntity, originalIqProtocolEntity):
        print(errorSyncIqProtocolEntity)
        raise KeyboardInterrupt()



import time
import logging
from threading import Thread, Lock
from yowsup.layers import YowProtocolLayer, YowLayerEvent, EventCallback
from yowsup.common import YowConstants
from yowsup.layers.network import YowNetworkLayer
from yowsup.layers.auth import YowAuthenticationProtocolLayer
from .protocolentities import *


class YowIqProtocolLayer(YowProtocolLayer):
    
    PROP_PING_INTERVAL               = "org.openwhatsapp.yowsup.prop.pinginterval"
    
    def __init__(self):
        handleMap = {
            "iq": (self.recvIq, self.sendIq)
        }
        self._pingThread = None
        self._pingQueue = {}
        self._pingQueueLock = Lock()
        self.__logger = logging.getLogger(__name__)
        super(YowIqProtocolLayer, self).__init__(handleMap)

    def __str__(self):
        return "Iq Layer"

    def onPong(self, protocolTreeNode, pingEntity):
        self.gotPong(pingEntity.getId())
        self.toUpper(ResultIqProtocolEntity.fromProtocolTreeNode(protocolTreeNode))

    def sendIq(self, entity):
        if entity.getXmlns() == "w:p":
            self._sendIq(entity, self.onPong)
        elif entity.getXmlns() in ("urn:xmpp:whatsapp:push", "w", "urn:xmpp:whatsapp:account", "encrypt"):
            self.toLower(entity.toProtocolTreeNode())

    def recvIq(self, node):
        if node["xmlns"] == "urn:xmpp:ping":
            entity = PongResultIqProtocolEntity(YowConstants.DOMAIN, node["id"])
            self.toLower(entity.toProtocolTreeNode())

    def gotPong(self, pingId):
        self._pingQueueLock.acquire()
        if pingId in self._pingQueue:
            self._pingQueue = {}
        self._pingQueueLock.release()

    def waitPong(self, id):
        self._pingQueueLock.acquire()
        self._pingQueue[id] = None
        pingQueueSize = len(self._pingQueue)
        self._pingQueueLock.release()
        self.__logger.debug("ping queue size: %d" % pingQueueSize)
        if pingQueueSize >= 2:
            self.getStack().broadcastEvent(YowLayerEvent(YowNetworkLayer.EVENT_STATE_DISCONNECT, reason = "Ping Timeout"))

    @EventCallback(YowAuthenticationProtocolLayer.EVENT_AUTHED)
    def onAuthed(self, event):
        interval = self.getProp(self.__class__.PROP_PING_INTERVAL, 50)
        if not self._pingThread and interval > 0:
            self._pingQueue = {}
            self._pingThread = YowPingThread(self, interval)
            self.__logger.debug("starting ping thread.")
            self._pingThread.start()
    
    
    def stop_thread(self):
        if self._pingThread:
            self.__logger.debug("stopping ping thread")
            if self._pingThread:
                self._pingThread.stop()
                self._pingThread = None
            self._pingQueue = {}
        
    @EventCallback(YowNetworkLayer.EVENT_STATE_DISCONNECT)
    def onDisconnect(self, event):
        self.stop_thread()
    
    @EventCallback(YowNetworkLayer.EVENT_STATE_DISCONNECTED)
    def onDisconnected(self, event):
        self.stop_thread()
            
class YowPingThread(Thread):
    def __init__(self, layer, interval):
        assert type(layer) is YowIqProtocolLayer, "layer must be a YowIqProtocolLayer, got %s instead." % type(layer)
        self._layer = layer
        self._interval = interval
        self._stop = False
        self.__logger = logging.getLogger(__name__)
        super(YowPingThread, self).__init__()
        self.daemon = True
        self.name = "YowPing%s" % self.name

    def run(self):
        while not self._stop:
            for i in range(0, self._interval):
                time.sleep(1)
                if self._stop:
                    self.__logger.debug("%s - ping thread stopped" % self.name)
                    return
            ping = PingIqProtocolEntity()
            self._layer.waitPong(ping.getId())
            if not self._stop:
                self._layer.sendIq(ping)

    def stop(self):
        self._stop = True



from yowsup.layers import YowLayer, YowLayerEvent, YowProtocolLayer
from .protocolentities import *
class YowIbProtocolLayer(YowProtocolLayer):

    def __init__(self):
        handleMap = {
            "ib": (self.recvIb, self.sendIb),
            "iq": (None, self.sendIb)
        }
        super(YowIbProtocolLayer, self).__init__(handleMap)

    def __str__(self):
        return "Ib Layer"

    def sendIb(self, entity):
        if entity.__class__ == CleanIqProtocolEntity:
            self.toLower(entity.toProtocolTreeNode())

    def recvIb(self, node):
        if node.getChild("dirty"):
            self.toUpper(DirtyIbProtocolEntity.fromProtocolTreeNode(node))
        elif node.getChild("offline"):
            self.toUpper(OfflineIbProtocolEntity.fromProtocolTreeNode(node))
        elif node.getChild("account"):
            self.toUpper(AccountIbProtocolEntity.fromProtocolTreeNode(node))
        else:
            raise ValueError("Unkown ib node %s" % node)




from yowsup.structs import ProtocolTreeNode
from yowsup.layers.protocol_iq.protocolentities import ResultIqProtocolEntity
class InfoGroupsResultIqProtocolEntity(ResultIqProtocolEntity):
    '''
<iq type="result" from="{{GROUP_ID}}" id="{{IQ_ID}}">
  <group subject="{{GROUPSUBJ}}" creation="{{GROUP_CREATION_TYIME}}"
      creator="{{CREATOR_JID}}" s_t="{{SUBJECT_SET_TIMESTAMP}}" id="{{GROUP_ID}}"
      s_o="{{SUBJECT_OWNER_JID}}">
    <participant jid="{{PARTICIPANT_JID}}" type="admin"></participant>
    <participant jid="{{PARTICIPANT_JID}}"></participant>
    <participant jid="{{PARTICIPANT_JID}}"></participant>
  </group>
</iq>
    '''
    TYPE_PARTICIPANT_ADMIN = "admin"
    def __init__(self, _id, _from,
                 groupId, creationTimestamp, creatorJid,
                 subject, subjectTime, subjectOwnerJid,
                 participants):
        super(InfoGroupsResultIqProtocolEntity, self).__init__(_id = _id, _from = _from)
        self.setGroupProps(groupId, creationTimestamp, creatorJid,
                           subject, subjectTime, subjectOwnerJid, participants)

    def setGroupProps(self, groupId, creationTimestamp, creatorJid,
                      subject, subjectTime, subjectOwnerJid,
                      participants):

        assert type(participants) is dict, "Participants must be a dict {jid => type?}"

        self.groupId = groupId
        self.creationTimestamp = int(creationTimestamp)
        self.creatorJid = creatorJid
        self.subject = subject
        self.subjectTime = int(subjectTime)
        self.subjectOwnerJid = subjectOwnerJid
        self.participants = participants

    def getParticipants(self):
        return self.participants

    def getSubject(self):
        return self.subject

    def getGroupId(self):
        return self.groupId

    def getCreationTimestamp(self):
        return self.creationTimestamp

    def getCreatorJid(self, full = True):
        return self.creatorJid if full else self.creatorJid.split('@')[0]

    def getSubjectTimestamp(self):
        return self.subjectTime

    def getSubjectOwnerJid(self, full = True):
        return self.subjectOwnerJid if full else self.subjectOwnerJid.split('@')[0]

    def getGroupAdmins(self, full = True):
        admins = []
        for jid, _type in self.participants.items():
            if _type == self.__class__.TYPE_PARTICIPANT_ADMIN:
                admins.append(jid if full else jid.split('@')[0])
        return admins

    def __str__(self):
        out = super(InfoGroupsResultIqProtocolEntity, self).__str__()
        out += "Group ID: %s\n" % self.groupId
        out += "Created: %s\n" % self.creationTimestamp
        out += "Creator JID: %s\n" % self.creatorJid
        out += "Subject: %s\n" % self.subject
        out += "Subject Timestamp: %s\n" % self.subjectTime
        out += "Subject owner JID: %s\n" % self.subjectOwnerJid
        out += "Participants: %s\n" % self.participants
        return out

    def toProtocolTreeNode(self):
        node = super(InfoGroupsResultIqProtocolEntity, self).toProtocolTreeNode()
        groupNode = ProtocolTreeNode("group", {
            "subject": self.getSubject(),
            "creation": str(self.getCreationTimestamp()),
            "creator": self.getCreatorJid(),
            "s_t": self.getSubjectTimestamp(),
            "s_o": self.getSubjectOwnerJid(),
            "id": self.getGroupId()
        })

        participants = []
        for jid, _type in self.getParticipants().items():
            pnode = ProtocolTreeNode("participant", {"jid": jid})
            if _type:
                pnode["type"] = _type
            participants.append(pnode)

        groupNode.addChildren(participants)
        node.addChild(groupNode)

        return node

    @staticmethod
    def fromProtocolTreeNode(node):
        groupNode = node.getChild("group")
        participants = {}
        for p in groupNode.getAllChildren("participant"):
            participants[p["jid"]] = p["type"]

        return InfoGroupsResultIqProtocolEntity(
            node["id"], node["from"],
            groupNode["id"], groupNode["creation"], groupNode["creator"], groupNode["subject"],
            groupNode["s_t"], groupNode["s_o"], participants
        )



from yowsup.structs import ProtocolEntity, ProtocolTreeNode
from .iq_result import ResultIqProtocolEntity
class PongResultIqProtocolEntity(ResultIqProtocolEntity):

    '''
    <iq type="result" xmlns="w:p" to="self.domain" id="1416174955-ping">
    </iq>
    '''
    def __init__(self, to, _id = None):
        super(PongResultIqProtocolEntity, self).__init__("w:p", _id = _id, to = to)



import threading, inspect, shlex
try:
    import Queue
except ImportError:
    import queue as Queue
try:
    import readline
except ImportError:
    import pyreadline as readline

class clicmd(object):
    def __init__(self, desc, order = 0):
        self.desc = desc
        self.order = order
    def __call__(self, fn):
        fn.clidesc = self.desc
        fn.cliorder = self.order
        return fn

class Cli(object):
    def __init__(self):
        self.sentCache = {}
        self.commands = {}
        self.acceptingInput = False
        self.lastPrompt = True
        self.blockingQueue = Queue.Queue()

        self._queuedCmds = []

        readline.set_completer(self.complete)
        readline.parse_and_bind('tab: complete')

        members = inspect.getmembers(self, predicate = inspect.ismethod)
        for m in members:
            if hasattr(m[1], "clidesc"):
                fname = m[0]
                fn = m[1]
                try:
                    cmd, subcommand = fname.split('_')
                except ValueError:
                    cmd = fname
                    subcommand = "_"


                if not cmd in self.commands:
                    self.commands[cmd] = {}

                self.commands[cmd][subcommand] = {
                   "args": inspect.getargspec(fn)[0][1:],
                   "optional": len(inspect.getargspec(fn)[3]) if inspect.getargspec(fn)[3] else 0,
                   "desc": fn.clidesc,
                   "fn": fn,
                   "order": fn.cliorder
                }
        #self.cv = threading.Condition()
        self.inputThread = threading.Thread(target = self.startInputThread)
        self.inputThread.daemon = True

    

    def queueCmd(self, cmd):
        self._queuedCmds.append(cmd)

    def startInput(self):
        self.inputThread.start()

    ################### cmd input parsing ####################
    def print_usage(self):
        line_width = 100
        
        outArr = []

        def addToOut(ind, cmd):
            if ind >= len(outArr):
                outArr.extend([None] * (ind - len(outArr) + 1))


            if outArr[ind] != None:
                for i in range(len(outArr) - 1, 0, -1):
                    if outArr[i] is None:
                        outArr[i] = outArr[ind]
                        outArr[ind] = cmd
                        return
                outArr.append(cmd)
            else:
                outArr[ind] = cmd

        for cmd, subcommands in self.commands.items():
            for subcmd, subcmdDetails in subcommands.items():
                out = ""
                out += ("/%s " % cmd).ljust(15)
                out += ("%s " % subcmd if subcmd != "_" else "").ljust(15)
                args = ("%s " % " ".join(["<%s>" % c for c in subcmdDetails["args"][0:len(subcmdDetails["args"])-subcmdDetails["optional"]]]))
                args += ("%s " % " ".join(["[%s]" % c for c in subcmdDetails["args"][len(subcmdDetails["args"])-subcmdDetails["optional"]:]]))
                out += args.ljust(30)
                out += subcmdDetails["desc"].ljust(20)
                addToOut(subcmdDetails["order"], out)

        print("----------------------------------------------")
        print("\n" . join(outArr))
        print("----------------------------------------------")

    def execCmd(self, cmdInput):
        cmdInput = cmdInput.rstrip()

        if not len(cmdInput) > 1:
            return

        if cmdInput.startswith("/"):
            cmdInput = cmdInput[1:]
        else:
            self.print_usage()
            return

        cmdInputDissect = [c for c in shlex.split(cmdInput) if c]

        cmd = cmdInputDissect[0]

        if not cmd in self.commands:
            return self.print_usage()

        cmdData = self.commands[cmd]
        if len(cmdData) == 1 and "_" in cmdData:
            subcmdData = cmdData["_"]
            args = cmdInputDissect[1:] if len(cmdInputDissect) > 1 else []
        else:
            args = cmdInputDissect[2:] if len(cmdInputDissect) > 2 else []
            subcmd = cmdInputDissect[1] if len(cmdInputDissect) > 1 else ""
            if subcmd not in cmdData:
                return self.print_usage()

            subcmdData = cmdData[subcmd]

        targetFn = subcmdData["fn"]
        if len(subcmdData["args"]) < len(args) or len(subcmdData["args"]) - subcmdData["optional"] > len(args):
            return self.print_usage()

        return self.doExecCmd(lambda :targetFn(*args))

    def doExecCmd(self, fn):
        return fn()


    def startInputThread(self):
        #cv.acquire()
        # Fix Python 2.x.
        global input
        try: input = raw_input
        except NameError: pass

        while(True):

            cmd = self._queuedCmds.pop(0) if len(self._queuedCmds) else input(self.getPrompt()).strip()
            wait = self.execCmd(cmd)
            if wait:
                self.acceptingInput = False
                self.blockingQueue.get(True)
                #cv.wait()
                #self.inputThread.wait()
            self.acceptingInput = True
        #cv.release()

    def getPrompt(self):
        return "[%s]:" % ("connected" if self.connected else "offline")

    def printPrompt(self):
        #return "Enter Message or command: (/%s)" % ", /".join(self.commandMappings)
        print(self.getPrompt())

    def output(self, message, tag = "general", prompt = True):
        if self.acceptingInput == True and self.lastPrompt is True:
            print("")


        self.lastPrompt = prompt

        if tag is not None:
            print("%s: %s" % (tag, message))
        else:
            print(message)
        if prompt:
            self.printPrompt()

    def complete(self, text, state):
        if state == 0:
            for cmd in self.commands:
                if cmd.startswith(text) and cmd != text:
                        return cmd

    def notifyInputThread(self):
        self.blockingQueue.put(1)

if __name__ == "__main__":
    c = Cli()
    c.print_usage()



from yowsup.structs import ProtocolEntity, ProtocolTreeNode
from yowsup.layers.protocol_receipts.protocolentities  import OutgoingReceiptProtocolEntity

class NotificationProtocolEntity(ProtocolEntity):
    '''
    <notification offline="0" id="{{NOTIFICATION_ID}}" notify="{{NOTIFY_NAME}}" type="{{NOTIFICATION_TYPE}}" 
            t="{{TIMESTAMP}}" from="{{SENDER_JID}}">
    </notification>

    
    '''
    def __init__(self, _type, _id, _from, timestamp, notify, offline):
        super(NotificationProtocolEntity, self).__init__("notification")
        self._type      = _type
        self._id        = _id
        self._from      =_from
        self.timestamp  = int(timestamp)
        self.notify     = notify
        self.offline    = offline == "1"
   

    def __str__(self):
        out = "Notification\n"
        out += "From: %s\n" % self.getFrom()
        out += "Type: %s\n" % self.getType()
        return out

    def getFrom(self, full = True):
        return self._from if full else self._from.split('@')[0]

    def getType(self):
        return self._type

    def getId(self):
        return self._id

    def getTimestamp(self):
        return self.timestamp

    def toProtocolTreeNode(self):
        attribs = {
            "t"         : str(self.timestamp),
            "from"      : self._from,
            "offline"   : "1" if self.offline else "0",
            "type"      : self._type,
            "id"        : self._id,
            "notify"    : self.notify
        }
       
        return self._createProtocolTreeNode(attribs, children = None, data = None)

    def ack(self):
        return OutgoingReceiptProtocolEntity(self.getId(), self.getFrom())

    @staticmethod
    def fromProtocolTreeNode(node):
        return NotificationProtocolEntity(
            node.getAttributeValue("type"), 
            node.getAttributeValue("id"),
            node.getAttributeValue("from"),
            node.getAttributeValue("t"),
            node.getAttributeValue("notify"),
            node.getAttributeValue("offline")
            )



from axolotl.state.identitykeystore import IdentityKeyStore
from axolotl.ecc.curve import Curve
from axolotl.identitykey import IdentityKey
from axolotl.util.keyhelper import KeyHelper
from axolotl.identitykeypair import IdentityKeyPair
from axolotl.ecc.djbec import *


class LiteIdentityKeyStore(IdentityKeyStore):
    def __init__(self, dbConn):
        """
        :type dbConn: Connection
        """
        self.dbConn = dbConn
        dbConn.execute("CREATE TABLE IF NOT EXISTS identities (_id INTEGER PRIMARY KEY AUTOINCREMENT,"
                       "recipient_id INTEGER UNIQUE,"
                       "registration_id INTEGER, public_key BLOB, private_key BLOB,"
                       "next_prekey_id INTEGER, timestamp INTEGER);")

        #identityKeyPairKeys = Curve.generateKeyPair()
        #self.identityKeyPair = IdentityKeyPair(IdentityKey(identityKeyPairKeys.getPublicKey()),
        #                                       identityKeyPairKeys.getPrivateKey())
        # self.localRegistrationId = KeyHelper.generateRegistrationId()

    def getIdentityKeyPair(self):
        q = "SELECT public_key, private_key FROM identities WHERE recipient_id = -1"
        c = self.dbConn.cursor()
        c.execute(q)
        result = c.fetchone()

        publicKey, privateKey = result
        return IdentityKeyPair(IdentityKey(DjbECPublicKey(publicKey[1:])), DjbECPrivateKey(privateKey))

    def getLocalRegistrationId(self):
        q = "SELECT registration_id FROM identities WHERE recipient_id = -1"
        c = self.dbConn.cursor()
        c.execute(q)
        result = c.fetchone()
        return result[0] if result else None


    def storeLocalData(self, registrationId, identityKeyPair):
        q = "INSERT INTO identities(recipient_id, registration_id, public_key, private_key) VALUES(-1, ?, ?, ?)"
        c = self.dbConn.cursor()
        c.execute(q, (registrationId, identityKeyPair.getPublicKey().getPublicKey().serialize(),
                      identityKeyPair.getPrivateKey().serialize()))

        self.dbConn.commit()

    def saveIdentity(self, recipientId, identityKey):
        q = "DELETE FROM identities WHERE recipient_id=?"
        self.dbConn.cursor().execute(q, (recipientId,))
        self.dbConn.commit()


        q = "INSERT INTO identities (recipient_id, public_key) VALUES(?, ?)"
        c = self.dbConn.cursor()
        c.execute(q, (recipientId, identityKey.getPublicKey().serialize()))
        self.dbConn.commit()

    def isTrustedIdentity(self, recipientId, identityKey):
        q = "SELECT public_key from identities WHERE recipient_id = ?"
        c = self.dbConn.cursor()
        c.execute(q, (recipientId,))
        result = c.fetchone()
        if not result:
            return True
        return result[0] == identityKey.getPublicKey().serialize()


from yowsup.layers import YowLayer
import logging
logger = logging.getLogger(__name__)
class YowLoggerLayer(YowLayer):

    def send(self, data):
        ldata = list(data) if type(data) is bytearray else data
        logger.debug("tx:\n%s" % ldata)
        self.toLower(data)

    def receive(self, data):
        ldata = list(data) if type(data) is bytearray else data
        logger.debug("rx:\n%s" % ldata)
        self.toUpper(data)

    def __str__(self):
        return "Logger Layer"


from yowsup.structs import ProtocolEntity, ProtocolTreeNode
class IbProtocolEntity(ProtocolEntity):
    '''
    <ib></ib>
    '''
    def __init__(self):
        super(IbProtocolEntity, self).__init__("ib")
    
    def toProtocolTreeNode(self):
        return self._createProtocolTreeNode({}, None, None)

    def __str__(self):
        out  = "Ib:\n"
        return out

    @staticmethod
    def fromProtocolTreeNode(node):
        return IbProtocolEntity()



from yowsup.structs import ProtocolEntity, ProtocolTreeNode
from yowsup.layers.protocol_iq.protocolentities import IqProtocolEntity
class CleanIqProtocolEntity(IqProtocolEntity):
    '''
    <iq id="" type="set" to="self.domain" xmlns="urn:xmpp:whatsapp:dirty">
        <clean type="{{dirty_type}}"></clean>
    </iq>
    '''
    def __init__(self, cleanType, to, _id = None):
        super(CleanIqProtocolEntity, self).__init__(
            "urn:xmpp:whatsapp:dirty",
            _id = _id,
            _type = "set",
            to = to
        )
        self.setProps(cleanType)

    def setProps(self, cleanType):
        self.cleanType = cleanType

    def __str__(self):
        out = super(CleanIqProtocolEntity, self).__str__()
        out += "Clean Type: %s\n" % self.cleanType
        return out

    def toProtocolTreeNode(self):
        node = super(CleanIqProtocolEntity, self).toProtocolTreeNode()
        cleanNode = ProtocolTreeNode("clean", {"type": self.cleanType})
        node.addChild(cleanNode)
        return node

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = IqProtocolEntity.fromProtocolTreeNode(node)
        entity.__class__ = CleanIqProtocolEntity
        entity.setProps(node.getChild("clean").getAttributeValue("type"))
        return entity


import binascii
import sys
class ProtocolTreeNode(object):
    def __init__(self, tag, attributes = None, children = None, data = None):

        self.tag = tag
        self.attributes = attributes or {}
        self.children = children or []
        self.data = data

        assert type(self.children) is list, "Children must be a list, got %s" % type(self.children)

    def __eq__(self, protocolTreeNode):
        """
        :param protocolTreeNode: ProtocolTreeNode
        :return: bool
        """
        #
        if protocolTreeNode.__class__ == ProtocolTreeNode\
            and self.tag == protocolTreeNode.tag\
            and self.data == protocolTreeNode.data\
            and self.attributes == protocolTreeNode.attributes\
            and len(self.getAllChildren()) == len(protocolTreeNode.getAllChildren()):
                found = False
                for c in self.getAllChildren():
                    for c2 in protocolTreeNode.getAllChildren():
                        if c == c2:
                            found = True
                            break
                    if not found:
                        return False

                found = False
                for c in protocolTreeNode.getAllChildren():
                    for c2 in self.getAllChildren():
                        if c == c2:
                            found = True
                            break
                    if not found:
                        return False

                return True

        return False

    def __hash__(self):
        return hash(self.tag) ^ hash(tuple(self.attributes.items())) ^ hash(self.data)

    def toString(self):
        out = "<"+self.tag
        if self.attributes is not None:
            for key,val in self.attributes.items():
                if val is None:
                    raise Exception("None val for key: "+key);
                out+= " "+key+'="'+val+'"'
        out+= ">\n"

        if self.data is not None:
            if type(self.data) is bytearray:
                try:
                    out += "%s" % self.data.decode()
                except UnicodeDecodeError:
                    out += binascii.hexlify(self.data)
            else:
                try:
                    out += "%s" % self.data
                except UnicodeDecodeError:
                    try:
                        out += "%s" % self.data.decode()
                    except UnicodeDecodeError:
                        out += binascii.hexlify(self.data)

            if type(self.data) is str and sys.version_info >= (3,0):
                out += "\nHEX3:%s\n" % binascii.hexlify(self.data.encode('latin-1'))
            else:
                out += "\nHEX:%s\n" % binascii.hexlify(self.data)
        
        for c in self.children:
            try:
                out += c.toString()
            except UnicodeDecodeError:
                out += "[ENCODED DATA]\n"
        out+= "</"+self.tag+">\n"
        return out

    
    def __str__(self):
        return self.toString() 

    def getData(self):
        return self.data

    def setData(self, data):
        self.data = data
        
    
    @staticmethod   
    def tagEquals(node,string):
        return node is not None and node.tag is not None and node.tag == string
        
        
    @staticmethod
    def require(node,string):
        if not ProtocolTreeNode.tagEquals(node,string):
            raise Exception("failed require. string: "+string);
    

    def __getitem__(self, key):
        return self.getAttributeValue(key)

    def __setitem__(self, key, val):
        self.setAttribute(key, val)

    def __delitem__(self, key):
        self.removeAttribute(key)


    def getChild(self,identifier):

        if type(identifier) == int:
            if len(self.children) > identifier:
                return self.children[identifier]
            else:
                return None

        for c in self.children:
            if identifier == c.tag:
                return c

        return None

    def hasChildren(self):
        return len(self.children) > 0

    def addChild(self, childNode):
        self.children.append(childNode)

    def addChildren(self, children):
        for c in children:
            self.addChild(c)
        
    def getAttributeValue(self,string):
        try:
            return self.attributes[string]
        except KeyError:
            return None

    def removeAttribute(self, key):
        if key in self.attributes:
            del self.attributes[key]

    def setAttribute(self, key, value):
        self.attributes[key] = value

    def getAllChildren(self,tag = None):
        ret = []
        if tag is None:
            return self.children
        
        for c in self.children:
            if tag == c.tag:
                ret.append(c)

        return ret



from yowsup.layers.auth.protocolentities.success import SuccessProtocolEntity
from yowsup.structs import ProtocolTreeNode
from yowsup.structs.protocolentity import ProtocolEntityTest
import unittest

class SuccessProtocolEntityTest(ProtocolEntityTest, unittest.TestCase):
    def setUp(self):
        self.ProtocolEntity = SuccessProtocolEntity
        attribs = {
            "status": "active",
            "kind": "free",
            "creation": "1234",
            "expiration": "1446578937",
            "props": "2",
            "t": "1415470561"
        }
        self.node = ProtocolTreeNode("success", attribs, None, "dummydata")


from yowsup.layers import YowLayer, YowLayerEvent, YowProtocolLayer
from .protocolentities import *
class YowReceiptProtocolLayer(YowProtocolLayer):
    def __init__(self):
        handleMap = {
            "receipt": (self.recvReceiptNode, self.sendReceiptEntity)
        }
        super(YowReceiptProtocolLayer, self).__init__(handleMap)

    def __str__(self):
        return "Receipt Layer"

    def sendReceiptEntity(self, entity):
        self.entityToLower(entity)

    def recvReceiptNode(self, node):
        self.toUpper(IncomingReceiptProtocolEntity.fromProtocolTreeNode(node))





from .message_media import MediaMessageProtocolEntity
from .message_media_downloadable import DownloadableMediaMessageProtocolEntity
from .message_media_downloadable_image import ImageDownloadableMediaMessageProtocolEntity
from .message_media_downloadable_audio import AudioDownloadableMediaMessageProtocolEntity
from .message_media_downloadable_video import VideoDownloadableMediaMessageProtocolEntity
from .message_media_location import LocationMediaMessageProtocolEntity
from .message_media_vcard import VCardMediaMessageProtocolEntity
from .iq_requestupload import RequestUploadIqProtocolEntity
from .iq_requestupload_result import ResultRequestUploadIqProtocolEntity



from .iq_picture import PictureIqProtocolEntity
from yowsup.structs import ProtocolTreeNode
import time
class SetPictureIqProtocolEntity(PictureIqProtocolEntity):
    '''
    <iq type="set" id="{{id}}" xmlns="w:profile:picture", to={{jid}}">
        <picture type="image" id="{{another_id}}">
        {{Binary bytes of the picture when type is set.}}
        </picture>
    </iq>
'''
    def __init__(self, jid, previewData, pictureData, pictureId = None, _id = None):
        super(SetPictureIqProtocolEntity, self).__init__(jid, _id, "set")
        self.setSetPictureProps(previewData, pictureData, pictureId)

    def setSetPictureProps(self, previewData, pictureData, pictureId = None):
        self.setPictureData(pictureData)
        self.setPictureId(pictureId or str(int(time.time())))
        self.setPreviewData(previewData)

    def setPictureData(self, pictureData):
        self.pictureData = pictureData

    def getPictureData(self):
        return self.pictureData

    def setPreviewData(self, previewData):
        self.previewData = previewData

    def getPreviewData(self):
        return self.previewData

    def setPictureId(self, pictureId):
        self.pictureId = pictureId

    def getPictureId(self):
        return self.pictureId

    def toProtocolTreeNode(self):
        node = super(PictureIqProtocolEntity, self).toProtocolTreeNode()
        attribs = {"type": "image", "id": self.pictureId}

        pictureNode = ProtocolTreeNode("picture", attribs, None, self.getPictureData())
        previewNode = ProtocolTreeNode("picture", {"type": "preview"}, None, self.getPreviewData())

        node.addChild(pictureNode)
        node.addChild(previewNode)

        return node

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = PictureIqProtocolEntity.fromProtocolTreeNode(node)
        entity.__class__ = SetPictureIqProtocolEntity

        pictureNode = None
        previewNode = None

        for child in node.getAllChildren("picture"):
            nodeType = child.getAttributeValue("type")
            if nodeType == "image":
                pictureNode = child
            elif nodeType == "preview":
                previewNode = child

        entity.setSetPictureProps(previewNode.getData(), pictureNode.getData(), pictureNode.getAttributeValue("id"))
        return entity


from yowsup.layers import YowLayer, YowLayerEvent, YowProtocolLayer
from .protocolentities import *
class YowPrivacyProtocolLayer(YowProtocolLayer):
    def __init__(self):
        handleMap = {
            "iq": (self.recvIq, self.sendIq)
        }
        super(YowPrivacyProtocolLayer, self).__init__(handleMap)

    def __str__(self):
        return "Privacy Layer"

    def sendIq(self, entity):
        if entity.getXmlns() == "jabber:iq:privacy":
            self.entityToLower(entity)

    def recvIq(self, node):
        pass



from yowsup.common import YowConstants
from yowsup.layers.protocol_iq.protocolentities import IqProtocolEntity
from yowsup.structs import ProtocolTreeNode
import os, time
class SetKeysIqProtocolEntity(IqProtocolEntity):
    def __init__(self, identityKey, signedPreKey, preKeys, djbType, registrationId = None):
        super(SetKeysIqProtocolEntity, self).__init__("encrypt", _type = "set", to = YowConstants.WHATSAPP_SERVER)
        self.setProps(identityKey, signedPreKey, preKeys, djbType, registrationId)

    def setProps(self, identityKey, signedPreKey, preKeys, djbType, registrationId = None):
        assert type(preKeys) is dict, "Expected keys to be a dict key_id -> public_key"
        assert type(signedPreKey) is tuple, "Exception signed pre key to be tuple id,key,signature"
        self.preKeys = preKeys
        self.identityKey = identityKey
        self.registration = registrationId or os.urandom(4)
        self.djbType = int(djbType)
        self.signedPreKey = signedPreKey

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = IqProtocolEntity.fromProtocolTreeNode(node)
        entity.__class__ = SetKeysIqProtocolEntity
        regVal = node.getChild("registration").data
        typeVal = node.getChild("type").data
        idVal = node.getChild("identity").data

        preKeys = {}
        for keyNode in node.getChild("list").getAllChildren():
            preKeys[keyNode.getChild("id").data] = keyNode.getChild("value").data


        skeyNode = node.getChild("skey")
        entity.setProps(idVal, (skeyNode.getChild("id").data, skeyNode.getChild("value").data,
                                skeyNode.getChild("signature").data), preKeys, typeVal, regVal)

        return entity

    def toProtocolTreeNode(self):
        node = super(SetKeysIqProtocolEntity, self).toProtocolTreeNode()
        identityNode = ProtocolTreeNode("identity", data = self.identityKey)

        listNode = ProtocolTreeNode("list")
        keyNodes = []
        for keyId, pk in self.preKeys.items():
            keyNode = ProtocolTreeNode("key")
            keyNode.addChild(ProtocolTreeNode("id", data = keyId))
            keyNode.addChild(ProtocolTreeNode("value", data = pk))
            keyNodes.append(keyNode)

        listNode.addChildren(keyNodes)

        regNode = ProtocolTreeNode("registration", data = self.registration)
        typeNode = ProtocolTreeNode("type", data = chr(self.djbType))
        _id, val, signature = self.signedPreKey
        skeyNode = ProtocolTreeNode("skey", children = [
            ProtocolTreeNode("id", data = _id),
            ProtocolTreeNode("value", data = val),
            ProtocolTreeNode("signature", data = signature)
        ])

        node.addChildren([
            listNode,
            identityNode,
            regNode,
            typeNode,
            skeyNode
        ])

        return node



from .iq_picture import PictureIqProtocolEntity
from yowsup.structs import ProtocolTreeNode
class GetPictureIqProtocolEntity(PictureIqProtocolEntity):
    '''
    <iq type="get" id="{{id}}" xmlns="w:profile:picture", to={{jid}}">
        <picture type="image | preview">
        </picture>
    </iq>'''
    def __init__(self, jid, preview = True, _id = None):
        super(GetPictureIqProtocolEntity, self).__init__(jid, _id, "get")
        self.setGetPictureProps(preview)

    def setGetPictureProps(self, preview = True):
        self.preview = preview

    def isPreview(self):
        return self.preview

    def toProtocolTreeNode(self):
        node = super(GetPictureIqProtocolEntity, self).toProtocolTreeNode()
        pictureNode = ProtocolTreeNode("picture", {"type": "preview" if self.isPreview() else "image" })
        node.addChild(pictureNode)
        return node

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = PictureIqProtocolEntity.fromProtocolTreeNode(node)
        entity.__class__ = GetPictureIqProtocolEntity
        entity.setGetPictureProps(node.getChild("picture").getAttributeValue("type"))
        return entity


from yowsup.layers import YowLayer, YowLayerEvent, YowProtocolLayer
from yowsup.common import YowConstants
from .protocolentities import *
class YowContactsIqProtocolLayer(YowProtocolLayer):
    def __init__(self):
        handleMap = {
            "iq": (self.recvIq, self.sendIq),
            "notification": (self.recvNotification, None)
        }
        super(YowContactsIqProtocolLayer, self).__init__(handleMap)

    def __str__(self):
        return "Iq Layer"

    def recvNotification(self, node):
        if node["type"] == "contacts":
            if node.getChild("remove"):
                self.toUpper(RemoveContactNotificationProtocolEntity.fromProtocolTreeNode(node))
            elif node.getChild("add"):
                self.toUpper(AddContactNotificationProtocolEntity.fromProtocolTreeNode(node))
            elif node.getChild("update"):
                self.toUpper(UpdateContactNotificationProtocolEntity.fromProtocolTreeNode(node))
            elif node.getChild("sync"):
                self.toUpper(ContactsSyncNotificationProtocolEntity.fromProtocolTreeNode(node))
            else:
                self.raiseErrorForNode(node)

    def recvIq(self, node):
        if node["type"] == "result" and node.getChild("sync"):
            self.toUpper(ResultSyncIqProtocolEntity.fromProtocolTreeNode(node))
        elif node["type"] == "result" and node.getChild("status"):
            self.toUpper(ResultStatusesIqProtocolEntity.fromProtocolTreeNode(node))

    def sendIq(self, entity):
        if entity.getXmlns() == "urn:xmpp:whatsapp:sync":
            self.toLower(entity.toProtocolTreeNode())
        elif entity.getXmlns() == GetStatusesIqProtocolEntity.XMLNS:
            self.toLower(entity.toProtocolTreeNode())



from yowsup.layers.protocol_presence.protocolentities.presence_unsubscribe import UnsubscribePresenceProtocolEntity
from yowsup.layers.protocol_presence.protocolentities.test_presence import PresenceProtocolEntityTest

class UnsubscribePresenceProtocolEntityTest(PresenceProtocolEntityTest):
    def setUp(self):
        super(UnsubscribePresenceProtocolEntityTest, self).setUp()
        self.ProtocolEntity = UnsubscribePresenceProtocolEntity
        self.node.setAttribute("type", "unsubscribe")
        self.node.setAttribute("to", "some_jid")


from yowsup.common import YowConstants
from yowsup.layers.protocol_iq.protocolentities import IqProtocolEntity
from yowsup.structs import ProtocolTreeNode

class UnregisterIqProtocolEntity(IqProtocolEntity):

    XMLNS = "urn:xmpp:whatsapp:account"

    def __init__(self):
        super(UnregisterIqProtocolEntity, self).__init__(_type = "get", to = YowConstants.WHATSAPP_SERVER)

    def toProtocolTreeNode(self):
        node = super(UnregisterIqProtocolEntity, self).toProtocolTreeNode()
        rmNode = ProtocolTreeNode("remove", {"xmlns": self.__class__.XMLNS})
        node.addChild(rmNode)
        return node

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = IqProtocolEntity.fromProtocolTreeNode(node)
        entity.__class__ = UnregisterIqProtocolEntity
        removeNode = node.getChild("remove")
        assert removeNode["xmlns"] == UnregisterIqProtocolEntity.XMLNS, "Not an account delete xmlns, got %s" % removeNode["xmlns"]

        return entity


from yowsup.common.http.warequest import WARequest
from yowsup.common.http.waresponseparser import JSONResponseParser
# from yowsup.env import CURRENT_ENV
from yowsup.common.tools import StorageTools, WATools
from yowsup.registration.existsrequest import WAExistsRequest
from yowsup.env import AndroidYowsupEnv
import random, hashlib, os
CURRENT_ENV = AndroidYowsupEnv()

class WACodeRequest(WARequest):

    def __init__(self,cc, p_in, mcc= "000", mnc = "000", sim_mcc = "000", sim_mnc = "000", method="sms"):
        super(WACodeRequest,self).__init__()
        idx = StorageTools.getIdentity(cc + p_in)

        self.p_in = p_in
        self.__id = idx
        self.cc = cc

        self.addParam("cc", cc)
        self.addParam("in", p_in)
        self.addParam("lc", "GB")
        self.addParam("lg", "en")
        self.addParam("sim_mcc", sim_mcc.zfill(3))
        self.addParam("sim_mnc", sim_mnc.zfill(3))
        self.addParam("mcc", sim_mcc.zfill(3))
        self.addParam("mnc", sim_mnc.zfill(3))
        self.addParam("method", method)

        self.addParam("mistyped", "6")
        self.addParam("network_radio_type", "1")
        self.addParam("simnum", "1")
        self.addParam("s", "")
        self.addParam("copiedrc", "1")
        self.addParam("hasinrc", "1")
        self.addParam("rcmatch", "1")
        self.addParam("pid", int(random.uniform(100,9999)))
        self.addParam("rchash", hashlib.sha256(os.urandom(20)).hexdigest())
        self.addParam("anhash", os.urandom(20))
        self.addParam("extexist", "1")
        self.addParam("extstate", "1")

        self.addParam("token", CURRENT_ENV.getToken(p_in))

        self.url = "v.whatsapp.net/v2/code"

        self.pvars = ["status","reason","length", "method", "retry_after", "code", "param"] +\
                    ["login", "pw", "type", "expiration", "kind", "price", "cost", "currency", "price_expiration"]

        self.setParser(JSONResponseParser())

    def send(self, parser = None):
        if self.__id is not None:
            request = WAExistsRequest(self.cc, self.p_in, self.__id)
            result = request.send()
            if result["status"] == "ok":
                return result

        self.__id = WATools.generateIdentity()
        self.addParam("id", self.__id)

        res = super(WACodeRequest, self).send(parser)
        if res["status"] == "sent":
            StorageTools.writeIdentity(self.cc + self.p_in, self.__id)
        return res




import importlib
import logging
logger = logging.getLogger(__name__)
from contextlib import contextmanager

class OptionalModule(object):
    def __init__(self, modulename, failMessage = None, require = False):
        self.modulename = modulename
        self.require = require
        self.failMessage = failMessage

    def __enter__(self):
        return self.importFn

    def importFn(self, what = None):
        imp = self.modulename if not what else ("%s.%s" % (self.modulename, what))
        return importlib.import_module(imp)

    def __exit__(self, exc_type, exc_val, exc_tb):
        if isinstance(exc_val, ImportError):
            failMessage = self.failMessage if self.failMessage is not None else ("%s import failed" % self.modulename)
            if failMessage:
                logger.error(failMessage)
            if self.require:
                raise
            return True

class PILOptionalModule(OptionalModule):
    def __init__(self, failMessage = None, require = False):
        super(PILOptionalModule, self).__init__("PIL",
                failMessage= failMessage,
                require = require)

class FFVideoOptionalModule(OptionalModule):
    def __init__(self, failMessage = None, require = False):
        super(FFVideoOptionalModule, self).__init__("ffvideo",
                failMessage=failMessage,
                require=require)

class AxolotlOptionalModule(OptionalModule):
    def __init__(self, failMessage = None, require = False):
        super(AxolotlOptionalModule, self).__init__("axolotl",
                failmessage=failMessage,
                require=require)

if __name__ == "__main__":
    with PILOptionalModule() as imp:
        imp()
        print("OK")
    print("DONE")



from yowsup.structs import ProtocolEntity, ProtocolTreeNode
class ChallengeProtocolEntity(ProtocolEntity):
    def __init__(self, nonce):
        super(ChallengeProtocolEntity, self).__init__("challenge")
        self.nonce = nonce
    
    def getNonce(self):
        return self.nonce

    def toProtocolTreeNode(self):
        #return self._createProtocolTreeNode({}, children = None, data = self.nonce)
        return self._createProtocolTreeNode({}, children = [], data = "".join(map(chr, self.nonce)))

    def __str__(self):
        out = "Challenge\n"
        out += "Nonce: %s\n" % self.nonce
        return out

    @staticmethod
    def fromProtocolTreeNode(node):
        nonce = list(map(ord,node.getData()))
        entity = ChallengeProtocolEntity(bytearray(nonce))
        return entity



from yowsup.structs import ProtocolEntity, ProtocolTreeNode
from .iq_groups_participants import ParticipantsGroupsIqProtocolEntity

class RemoveParticipantsIqProtocolEntity(ParticipantsGroupsIqProtocolEntity):
    '''
    <iq type="set" id="{{id}}" xmlns="w:g2", to="{{group_jid}}">
        <remove>
            <participant jid="{{jid}}"></participant>
            <participant jid="{{jid}}"></participant>
        </remove>
    </iq>
    '''
    
    def __init__(self, group_jid, participantList, _id = None):
        super(RemoveParticipantsIqProtocolEntity, self).__init__(group_jid, participantList, "remove", _id = _id)
        
    @staticmethod
    def fromProtocolTreeNode(node):
        entity = super(RemoveParticipantsIqProtocolEntity, RemoveParticipantsIqProtocolEntity).fromProtocolTreeNode(node)
        entity.__class__ = RemoveParticipantsIqProtocolEntity
        participantList = []
        for participantNode in node.getChild("remove").getAllChildren():
            participantList.append(participantNode["jid"])
        entity.setProps(node.getAttributeValue("to"), participantList)
        return entity



from yowsup.common import YowConstants
from yowsup.layers.protocol_iq.protocolentities import IqProtocolEntity
from yowsup.structs import ProtocolTreeNode

class GetStatusesIqProtocolEntity(IqProtocolEntity):
    XMLNS = "status"

    def __init__(self, jids, _id = None):
        """
        Request the statuses of users. Should be sent once after login.

        Args:
            - jids: A list of jids representing the users whose statuses you are
                trying to get.
        """
        super(GetStatusesIqProtocolEntity, self).__init__(self.__class__.XMLNS, _id, _type = "get", to = YowConstants.WHATSAPP_SERVER)
        self.setGetStatusesProps(jids)

    def setGetStatusesProps(self, jids):
        assert type(jids) is list, "jids must be a list of jids"
        self.jids = jids

    def __str__(self):
        out = super(GetStatusesIqProtocolEntity, self).__str__()
        out += "Numbers: %s\n" % (",".join(self.numbers))
        return out

    def toProtocolTreeNode(self):
        users = [ProtocolTreeNode("user", {'jid': jid}) for jid in self.jids]

        node = super(GetStatusesIqProtocolEntity, self).toProtocolTreeNode()
        statusNode = ProtocolTreeNode("status", None, users)
        node.addChild(statusNode)

        return node

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = IqProtocolEntity.fromProtocolTreeNode(node)
        entity.__class__ = GetStatusesIqProtocolEntity
        statusNode = node.getChild("status")
        userNodes = statusNode.getAllChildren()
        jids = [user['jid'] for user in userNodes]

        entity.setGetStatusesProps(jids)

        return entity



from yowsup.layers.protocol_iq.protocolentities.iq import IqProtocolEntity
from yowsup.structs import ProtocolTreeNode
from yowsup.structs.protocolentity import ProtocolEntityTest
import unittest

class IqProtocolEntityTest(unittest.TestCase, ProtocolEntityTest):
    def setUp(self):
        self.ProtocolEntity = IqProtocolEntity
        self.node = ProtocolTreeNode("iq", {"id": "test_id", "type": "get", "xmlns": "iq_xmlns"}, None, None)


from yowsup.common import YowConstants
from yowsup.structs import ProtocolEntity, ProtocolTreeNode
from yowsup.layers.protocol_iq.protocolentities import ResultIqProtocolEntity
from ..structs import Group
class ListGroupsResultIqProtocolEntity(ResultIqProtocolEntity):
    '''
    <iq type="result" from="g.us" id="{{IQ_ID}}">
      <groups>
          <group s_t="{{SUBJECT_TIME}}" creation="{{CREATING_TIME}}" creator="{{OWNER_JID}}" id="{{GROUP_ID}}" s_o="{{SUBJECT_OWNER_JID}}" subject="{{SUBJECT}}">
            <participant jid="{{JID}}" type="admin">
            </participant>
            <participant jid="{{JID}}">
            </participant>
          </group>
          <group s_t="{{SUBJECT_TIME}}" creation="{{CREATING_TIME}}" creator="{{OWNER_JID}}" id="{{GROUP_ID}}" s_o="{{SUBJECT_OWNER_JID}}" subject="{{SUBJECT}}">
            <participant jid="{{JID}}" type="admin">
            </participant>
          </group>
      <groups>
    </iq>
    '''

    def __init__(self, groupsList):
        super(ListGroupsResultIqProtocolEntity, self).__init__(_from = YowConstants.WHATSAPP_GROUP_SERVER)
        self.setProps(groupsList)

    def __str__(self):
        out = super(ListGroupsResultIqProtocolEntity, self).__str__()
        out += "Groups:\n"
        for g in self.groupsList:
            out += "%s\n" % g
        return out

    def getGroups(self):
        return self.groupsList

    def setProps(self, groupsList):
        assert type(groupsList) is list and (len(groupsList) == 0 or groupsList[0].__class__ is Group),\
            "groupList must be a list of Group instances"
        self.groupsList = groupsList


    def toProtocolTreeNode(self):
        node = super(ListGroupsResultIqProtocolEntity, self).toProtocolTreeNode()

        groupsNodes = []
        for group in self.groupsList:
            groupNode = ProtocolTreeNode("group", {
                "id":       group.getId(),
                "creator":    group.getCreator(),
                "subject":  group.getSubject(),
                "s_o":      group.getSubjectOwner(),
                "s_t":      str(group.getSubjectTime()),
                "creation": str(group.getCreationTime())
                },
            )
            participants = []
            for jid, _type in group.getParticipants().items():
                pnode = ProtocolTreeNode("participant", {"jid": jid})
                if _type:
                    pnode["type"] = _type
                participants.append(pnode)
            groupNode.addChildren(participants)
            groupsNodes.append(groupNode)

        node.addChild(ProtocolTreeNode("groups", children = groupsNodes))
        return node

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = super(ListGroupsResultIqProtocolEntity, ListGroupsResultIqProtocolEntity).fromProtocolTreeNode(node)
        entity.__class__ = ListGroupsResultIqProtocolEntity
        groups = []
        for groupNode in node.getChild("groups").getAllChildren():
            participants = {}
            for p in groupNode.getAllChildren("participant"):
                participants[p["jid"]] = p["type"]
            groups.append(
                Group(groupNode["id"], groupNode["creator"], groupNode["subject"], groupNode["s_o"], groupNode["s_t"], groupNode["creation"], participants)
            )

        entity.setProps(groups)
        return entity



from yowsup.layers import YowLayer
from yowsup import ProtocolTreeNode
from .mediadownloader import MediaDownloader
import shutil, os, logging
logger = logging.getLogger(__name__)

class YowMediaPictureLayer(YowLayer):
    def send(self, data):
        self.toLower(data)

    def receive(self, node):
        if ProtocolTreeNode.tagEquals(node, "message") and node.getAttributeValue("type") == "media":
            self.downloadMedia(node.getChild("media").getAttributeValue("url"))
        else:
            self.toUpper(node)

    def downloadMedia(self, url):
        logger.debug("Downloading %s" % url)
        downloader = MediaDownloader(self.onSuccess, self.onError, self.onProgress)
        downloader.download(url)

    def onError(self):
        logger.error("Error download file")

    def onSuccess(self, path):
        outPath = "/tmp/yowfiles/%s.jpg" % os.path.basename(path)
        shutil.copyfile(path, outPath)
        logger.debug("Picture downloaded to %s" % outPath)

    def onProgress(self, progress):
        logger.debug("Download progress %s" % progress)



from yowsup.structs import ProtocolEntity, ProtocolTreeNode
from .presence import PresenceProtocolEntity
class AvailablePresenceProtocolEntity(PresenceProtocolEntity):
    '''
    <presence type="available"></presence>
    response:
    <presence from="self.jid">
    </presence>

    '''
    def __init__(self):
        super(AvailablePresenceProtocolEntity, self).__init__("available")


from yowsup.layers.protocol_messages.protocolentities import MessageProtocolEntity
from yowsup.structs import ProtocolTreeNode
import sys
class EncryptedMessageProtocolEntity(MessageProtocolEntity):
    '''
    <message retry="1" from="49xxxxxxxx@s.whatsapp.net" t="1418906418" offline="1" type="text" id="1418906377-1" notify="Tarek Galal">
<enc type="{{type}}" v="{{1 || 2}}">
HEX:33089eb3c90312210510e0196be72fe65913c6a84e75a54f40a3ee290574d6a23f408df990e718da761a210521f1a3f3d5cb87fde19fadf618d3001b64941715efd3e0f36bba48c23b08c82f2242330a21059b0ce2c4720ec79719ba862ee3cda6d6332746d05689af13aabf43ea1c8d747f100018002210d31cd6ebea79e441c4935f72398c772e2ee21447eb675cfa28b99de8d2013000</enc>
</message>
    '''

    TYPE_PKMSG = "pkmsg"
    TYPE_MSG = "msg"

    def __init__(self, encType, encVersion, encData, _type, _id = None,  _from = None, to = None, notify = None, timestamp = None,
            participant = None, offline = None, retry = None ):
        super(EncryptedMessageProtocolEntity, self).__init__(_type, _id = _id,  _from = _from, to = to, notify = notify,
                                                      timestamp = timestamp, participant = participant, offline = offline,
                                                      retry = retry)

        self.setEncProps(encType, encVersion, encData)

    def setEncProps(self, encType, encVersion, encData):
        assert encType in "pkmsg", "msg"
        self.encType = encType
        self.encVersion = int(encVersion)
        self.encData = encData

    def getEncType(self):
        return self.encType

    def getEncData(self):
        return self.encData

    def getVersion(self):
        return self.encVersion

    def toProtocolTreeNode(self):
        node = super(EncryptedMessageProtocolEntity, self).toProtocolTreeNode()
        encNode = ProtocolTreeNode("enc", data = self.encData)
        encNode["type"] = self.encType
        encNode["v"] = str(self.encVersion)

        node.addChild(encNode)
        return node

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = MessageProtocolEntity.fromProtocolTreeNode(node)
        entity.__class__ = EncryptedMessageProtocolEntity
        encNode = node.getChild("enc")
        entity.setEncProps(encNode["type"], encNode["v"],
                           encNode.data.encode('latin-1') if sys.version_info >= (3,0) else encNode.data)
        return entity


'''
Copyright (c) <2012> Tarek Galal <tare2.galal@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy of this 
software and associated documentation files (the "Software"), to deal in the Software 
without restriction, including without limitation the rights to use, copy, modify, 
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to 
permit persons to whom the Software is furnished to do so, subject to the following 
conditions:

The above copyright notice and this permission notice shall be included in all 
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR 
A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF 
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
'''

from yowsup.common.http.warequest import WARequest
from yowsup.common.http.waresponseparser import JSONResponseParser
from yowsup.common.tools import StorageTools

class WARegRequest(WARequest):

    def __init__(self,cc, p_in, code):
        super(WARegRequest,self).__init__()
        idx = StorageTools.getIdentity(cc + p_in)

        if idx is None:
            raise ValueError("You have to request code first")

        self.addParam("cc", cc)
        self.addParam("in", p_in)

        self.addParam("id", idx)
        self.addParam("code", code)

        self.url = "v.whatsapp.net/v2/register"

        self.pvars = ["status", "login", "pw", "type", "expiration", "kind", "price", "cost", "currency", "price_expiration",
                      "reason","retry_after"]

        self.setParser(JSONResponseParser())
        
    def register(self):
        return self.send()


from yowsup.layers import YowProtocolLayerTest
from yowsup.layers.protocol_chatstate import YowChatstateProtocolLayer
from yowsup.layers.protocol_chatstate.protocolentities import IncomingChatstateProtocolEntity, OutgoingChatstateProtocolEntity

class YowChatStateProtocolLayerTest(YowProtocolLayerTest, YowChatstateProtocolLayer):
    def setUp(self):
        YowChatstateProtocolLayer.__init__(self)

    def test_send(self):
        entity = OutgoingChatstateProtocolEntity(OutgoingChatstateProtocolEntity.STATE_PAUSED, "jid@s.whatsapp.net")
        self.assertSent(entity)

    def test_receive(self):
        entity = IncomingChatstateProtocolEntity(IncomingChatstateProtocolEntity.STATE_TYPING, "jid@s.whatsapp.net")
        self.assertReceived(entity)


from yowsup.layers import YowLayerInterface

class YowAuthenticationProtocolLayerInterface(YowLayerInterface):
    def setCredentials(self, phone, password):
        self._layer.setCredentials((phone, password))

    def getUsername(self, full = False):
        return self._layer.getUsername(full)
